"""Commands to finish up after a demultiplex run"""
import logging

import click

from cg.constants.demultiplexing import OPTION_BCL_CONVERTER
from cg.meta.demultiplex.demux_post_processing import DemuxPostProcessingAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.group(name="finish")
def finish_group():
    """Finish up after demultiplexing"""


@finish_group.command(name="all")
@OPTION_BCL_CONVERTER
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def finish_all_cmd(context: CGConfig, bcl_converter: str, dry_run: bool):
    """Command to post process all demultiplexed flowcells"""
    demux_post_processing_api = DemuxPostProcessingAPI(config=context)
    demux_post_processing_api.set_dry_run(dry_run=dry_run)
    demux_post_processing_api.finish_all_flowcells(bcl_converter=bcl_converter)


@finish_group.command(name="flowcell")
@click.argument("flowcell-name")
@OPTION_BCL_CONVERTER
@click.option("--dry-run", is_flag=True)
@click.option("--force", is_flag=True)
@click.pass_obj
def finish_flowcell(
    context: CGConfig, flowcell_name: str, bcl_converter: str, dry_run: bool, force: bool
):
    """Command to finish up a flowcell after demultiplexing

    flowcell-name is full flowcell name, e.g. '201203_A00689_0200_AHVKJCDRXX'
    """

    demux_post_processing_api = DemuxPostProcessingAPI(config=context)
    demux_post_processing_api.set_dry_run(dry_run)
    demux_post_processing_api.finish_flowcell(
        flowcell_name=flowcell_name, force=force, bcl_converter=bcl_converter
    )



"""Report helper"""


def recursive_assert(data):
    """Dictionary recursive assert test"""

    for k, v in data.items():
        if isinstance(v, dict):
            recursive_assert(dict(v))
        if isinstance(v, list):
            for e in v:
                recursive_assert(dict(e))
        else:
            assert v, f"{k} value is missing"



"""Tests for reset part of the store API"""

from typing import List

from cg.store import Store, models
from cg.store.api.import_func import (
    add_application_version,
    import_application_versions,
    import_applications,
    import_apptags,
    parse_application_versions,
    prices_are_same,
    versions_are_same,
)
from cg.store.api.models import ApplicationVersionSchema
from tests.store.api.conftest import StoreCheckers
from tests.store_helpers import StoreHelpers


def test_prices_are_same_int_and_int():
    # GIVEN float price that looks like one in excel
    # GIVEN same price but as it looks when saved in database
    float_price = 0.0
    database_price = 0

    # WHEN calling prices are same
    should_be_same = prices_are_same(database_price, float_price)

    # THEN prices should be considered same
    assert should_be_same


def test_prices_are_not_same_int_and_int():
    # GIVEN float price that looks like one in excel
    # GIVEN same price but as it looks when saved in database
    float_price = 1.0
    database_price = 0

    # WHEN calling prices are same
    should_not_be_same = prices_are_same(database_price, float_price)

    # THEN prices should be considered same
    assert not should_not_be_same


def test_prices_are_same_float_and_int():
    # GIVEN float price that looks like one in excel
    # GIVEN same price but as it looks when saved in database
    float_price = 0.654345
    database_price = 1

    # WHEN calling prices are same
    should_be_same = prices_are_same(database_price, float_price)

    # THEN prices should be considered same
    assert should_be_same


def test_prices_are_same_float_and_float():
    # GIVEN float price that looks like one in excel
    # GIVEN same price but as it looks when saved in database
    float_price = 0.654345
    database_price = float_price

    # WHEN calling prices are same
    should_be_same = prices_are_same(database_price, float_price)

    # THEN prices should be considered same
    assert should_be_same


def test_versions_are_same(applications_store: Store, application_versions_file: str):
    # GIVEN a database with some applications loaded
    store = applications_store

    # GIVEN an excel price row
    # same price row committed to the database
    excel_versions: List[ApplicationVersionSchema] = list(
        parse_application_versions(excel_path=application_versions_file)
    )
    version: ApplicationVersionSchema = excel_versions[0]

    # GIVEN that the application exists in the database
    application_obj: models.Application = store.application(version.app_tag)

    # GIVEN that there is a application version
    sign = "DummySign"
    db_version: models.ApplicationVersion = add_application_version(
        application_obj=application_obj,
        latest_version=None,
        version=version,
        sign=sign,
        store=store,
    )

    # WHEN calling versions are same
    should_be_same = versions_are_same(version_obj=db_version, application_version=version)

    # THEN versions are considered same
    assert should_be_same


def test_versions_are_not_same(applications_store: Store, application_versions_file: str):
    # GIVEN a database with some applications loaded

    # GIVEN an excel price row
    # NOT same price row committed to the database
    excel_versions: List[ApplicationVersionSchema] = list(
        parse_application_versions(excel_path=application_versions_file)
    )
    version: ApplicationVersionSchema = excel_versions[0]
    application_obj: models.Application = applications_store.application(version.app_tag)
    sign = "DummySign"

    db_version: models.ApplicationVersion = add_application_version(
        application_obj=application_obj,
        latest_version=None,
        version=version,
        sign=sign,
        store=applications_store,
    )

    another_version: ApplicationVersionSchema = excel_versions[1]

    # WHEN calling versions are same
    should_not_be_same: bool = versions_are_same(
        version_obj=db_version, application_version=another_version
    )

    # THEN versions are not considered same
    assert should_not_be_same is False


def test_application_version(
    applications_store: Store,
    application_versions_file: str,
    store: Store,
    store_checkers: StoreCheckers,
):
    # GIVEN a store with applications
    # and an excel file with prices for those applications
    sign = "TestSign"

    # WHEN calling import_application_version
    import_application_versions(
        store=applications_store,
        excel_path=application_versions_file,
        sign=sign,
        dry_run=False,
        skip_missing=False,
    )

    # THEN versions should have been created in the store
    assert store_checkers.all_versions_exist_in_store(
        store=applications_store, excel_path=application_versions_file
    )


def test_application_version_dry_run(
    applications_store: Store, application_versions_file: str, store_checkers: StoreCheckers
):
    # GIVEN a store with applications
    # and an excel file with prices for those applications
    sign = "TestSign"

    # WHEN calling import_application_version as dry run
    import_application_versions(
        store=applications_store,
        excel_path=application_versions_file,
        sign=sign,
        dry_run=True,
        skip_missing=False,
    )

    # THEN versions should not have been created in the store
    assert not store_checkers.all_versions_exist_in_store(
        store=applications_store, excel_path=application_versions_file
    )


def test_application(store: Store, applications_file: str, store_checkers: StoreCheckers):
    # GIVEN an excel file with applications
    assert store_checkers.all_applications_exists(store, applications_file) is False
    # GIVEN a store without the applications
    sign = "TestSign"

    # WHEN calling import_applications
    import_applications(store=store, excel_path=applications_file, sign=sign, dry_run=False)

    # THEN applications should have been created in the store
    assert store_checkers.all_applications_exists(store, applications_file)


def test_application_sheet_name(
    applications_file: str, store: Store, store_checkers: StoreCheckers
):
    # GIVEN a store and an excel file with applications
    sign = "TestSign"

    # WHEN calling import_applications
    import_applications(
        store=store,
        excel_path=applications_file,
        sign=sign,
        dry_run=False,
        sheet_name="application",
    )

    # THEN applications should have been created in the store
    assert store_checkers.all_applications_exists(store, applications_file)


def test_application_dry_run(applications_file: str, store: Store, store_checkers: StoreCheckers):
    # GIVEN a store and an excel file with applications
    sign = "TestSign"

    # WHEN calling import_applications as dry run
    import_applications(store=store, excel_path=applications_file, sign=sign, dry_run=True)

    # THEN applications should not have been created in the store
    assert not store_checkers.all_applications_exists(store, applications_file)


def test_sync_microbial_orderform_dry_run(microbial_store: Store, microbial_orderform: str):
    # GIVEN a microbial orderform and a store where all the apptags exists half some inactive and
    # some active
    prep_category = "mic"
    sign = "PG"
    activate = False
    inactivate = False
    active_mic_apps_from_start = microbial_store.applications(
        category=prep_category, archived=False
    ).count()
    inactive_mic_apps_from_start = microbial_store.applications(
        category=prep_category, archived=True
    ).count()

    # WHEN syncing app-tags in that orderform
    import_apptags(
        store=microbial_store,
        excel_path=microbial_orderform,
        prep_category=prep_category,
        signature=sign,
        activate=activate,
        inactivate=inactivate,
        sheet_name="Drop down list",
        tag_column=2,
        tag_row=1,
    )

    # THEN same number of active and inactive mic applications in status database
    active_mic_apps_after_when = microbial_store.applications(
        category=prep_category, archived=False
    ).count()
    inactive_mic_apps_after_when = microbial_store.applications(
        category=prep_category, archived=True
    ).count()
    assert active_mic_apps_from_start == active_mic_apps_after_when
    assert inactive_mic_apps_from_start == inactive_mic_apps_after_when


def test_sync_microbial_orderform_activate(microbial_store: Store, microbial_orderform: str):
    # GIVEN a microbial orderform and a store where all the apptags exists half some inactive and
    # some active
    prep_category = "mic"
    sign = "PG"
    activate = True
    inactivate = False
    active_mic_apps_from_start = microbial_store.applications(
        category=prep_category, archived=False
    ).count()
    inactive_mic_apps_from_start = microbial_store.applications(
        category=prep_category, archived=True
    ).count()

    # WHEN syncing app-tags in that orderform
    import_apptags(
        store=microbial_store,
        excel_path=microbial_orderform,
        prep_category=prep_category,
        signature=sign,
        activate=activate,
        inactivate=inactivate,
        sheet_name="Drop down list",
        tag_column=3,
        tag_row=1,
    )

    # THEN more active mic applications in status database and same inactive
    active_mic_apps_after_when = microbial_store.applications(
        category=prep_category, archived=False
    ).count()
    inactive_mic_apps_after_when = microbial_store.applications(
        category=prep_category, archived=True
    ).count()
    assert active_mic_apps_from_start < active_mic_apps_after_when
    assert inactive_mic_apps_from_start > inactive_mic_apps_after_when


def test_sync_rml_orderform_inactivate(rml_store: Store, rml_orderform: str, helpers: StoreHelpers):
    # GIVEN a rml orderform and a store where all the apptags exists half some inactive and
    # some active
    prep_category = "rml"
    sign = "PG"
    activate = False
    inactivate = True

    helpers.add_application(
        store=rml_store,
        application_tag="apptag_to_inactivate",
        is_archived=False,
        application_type=prep_category,
    )
    active_apps_from_start = rml_store.applications(category=prep_category, archived=False).count()
    inactive_apps_from_start = rml_store.applications(category=prep_category, archived=True).count()

    # WHEN syncing app-tags in that orderform
    import_apptags(
        store=rml_store,
        excel_path=rml_orderform,
        prep_category=prep_category,
        signature=sign,
        activate=activate,
        inactivate=inactivate,
        sheet_name="Drop down list",
        tag_column=0,
        tag_row=1,
    )

    # THEN the number of active should be less and the number of inactive more than before
    active_apps_after_when = rml_store.applications(category=prep_category, archived=False).count()
    inactive_apps_after_when = rml_store.applications(category=prep_category, archived=True).count()
    assert active_apps_from_start > active_apps_after_when
    assert inactive_apps_from_start < inactive_apps_after_when



from typing import Optional, Union

from pydantic import BaseModel, validator
from cg.models.report.validators import validate_empty_field, validate_float


class SampleMetadataModel(BaseModel):
    """
    Metrics and trending data model associated to a specific sample

    Attributes:
        million_read_pairs: number of million read pairs obtained; source: StatusDB/sample/reads (/2*10^6)
        duplicates: fraction of mapped sequence that is marked as duplicate; source: pipeline workflow

    """

    million_read_pairs: Union[None, float, str]
    duplicates: Union[None, float, str]

    _float_values = validator("million_read_pairs", "duplicates", always=True, allow_reuse=True)(
        validate_float
    )


class MipDNASampleMetadataModel(SampleMetadataModel):
    """Metrics and trending data model associated to a specific MIP DNA sample

    Attributes:
        bait_set: panel bed used for the analysis; source: LIMS
        gender: gender estimated by the pipeline; source: pipeline workflow
        mapped_reads: percentage of reads aligned to the reference sequence; source: pipeline workflow
        mean_target_coverage: mean coverage of a target region; source: pipeline workflow
        pct_10x: percent of targeted bases that are covered to 10X coverage or more; source: pipeline workflow
    """

    bait_set: Optional[str]
    gender: Optional[str]
    mapped_reads: Union[None, float, str]
    mean_target_coverage: Union[None, float, str]
    pct_10x: Union[None, float, str]

    _str_values_mip = validator("bait_set", "gender", always=True, allow_reuse=True)(
        validate_empty_field
    )

    _float_values_mip = validator(
        "mapped_reads", "mean_target_coverage", "pct_10x", always=True, allow_reuse=True
    )(validate_float)


class BalsamicSampleMetadataModel(SampleMetadataModel):
    """Metrics and trending data model associated to a specific BALSAMIC sample

    Attributes:
            mean_insert_size: mean insert size of the distribution; source: pipeline workflow
            fold_80: fold 80 base penalty; source: pipeline workflow
    """

    mean_insert_size: Union[None, float, str]
    fold_80: Union[None, float, str]

    _float_values_balsamic = validator(
        "mean_insert_size", "fold_80", always=True, allow_reuse=True
    )(validate_float)


class BalsamicTargetedSampleMetadataModel(BalsamicSampleMetadataModel):
    """Metrics and trending data model associated to a specific BALSAMIC sample

    Attributes:
            bait_set: panel bed used for the analysis; source: LIMS
            bait_set_version: panel bed version; source: pipeline workflow
            median_target_coverage: median coverage of a target region in bases; source: pipeline workflow
            pct_250x: percent of targeted bases that are covered to 250X coverage or more; source: pipeline workflow
            pct_500x: percent of targeted bases that are covered to 500X coverage or more; source: pipeline workflow
    """

    bait_set: Optional[str]
    bait_set_version: Union[None, int, str]
    median_target_coverage: Union[None, float, str]
    pct_250x: Union[None, float, str]
    pct_500x: Union[None, float, str]

    _str_values = validator("bait_set", "bait_set_version", always=True, allow_reuse=True)(
        validate_empty_field
    )

    _float_values_balsamic_targeted = validator(
        "median_target_coverage", "pct_250x", "pct_500x", always=True, allow_reuse=True
    )(validate_float)


class BalsamicWGSSampleMetadataModel(BalsamicSampleMetadataModel):
    """Metrics and trending data model associated to a specific BALSAMIC sample

    Attributes:
            median_coverage: median coverage in bases of the genome territory; source: pipeline workflow
            pct_15x: fraction of bases that attained at least 15X sequence coverage; source: pipeline workflow
            pct_60x: fraction of bases that attained at least 15X sequence coverage; source: pipeline workflow
    """

    median_coverage: Union[None, float, str]
    pct_15x: Union[None, float, str]
    pct_60x: Union[None, float, str]

    _float_values_balsamic_wgs = validator(
        "median_coverage", "pct_15x", "pct_60x", always=True, allow_reuse=True
    )(validate_float)



"""Test for analysis"""

import pytest

from cg.constants import Priority
from cg.constants.priority import SlurmQos
from cg.meta.workflow.analysis import AnalysisAPI


@pytest.mark.parametrize(
    "priority,expected_slurm_qos",
    [
        (Priority.clinical_trials, SlurmQos.NORMAL),
        (Priority.research, SlurmQos.LOW),
        (Priority.standard, SlurmQos.NORMAL),
        (Priority.priority, SlurmQos.HIGH),
        (Priority.express, SlurmQos.EXPRESS),
    ],
)
def test_get_slurm_qos_for_case(mocker, case_id: str, priority, expected_slurm_qos):
    """Test qet Quality of service (SLURM QOS) from the case priority"""

    # GIVEN a case that has a priority
    mocker.patch.object(AnalysisAPI, "get_priority_for_case")
    AnalysisAPI.get_priority_for_case.return_value = priority

    # When getting the SLURM QOS for the priority
    slurm_qos = AnalysisAPI.get_slurm_qos_for_case(AnalysisAPI, case_id=case_id)

    # THEN the expected slurm QOS should be returned
    assert slurm_qos is expected_slurm_qos



"""
    API for uploading observations
"""
from datetime import datetime
from pathlib import Path

import logging
from typing import List

from alchy import Query

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.loqus import LoqusdbAPI
from cg.constants.loqus_upload import ObservationAnalysisTag, ObservationAnalysisType
from cg.exc import CaseNotFoundError, DuplicateRecordError, DuplicateSampleError
from cg.models.observations.observations_input_files import ObservationsInputFiles
from cg.store import Store, models

LOG = logging.getLogger(__name__)


class UploadObservationsAPI:

    """API to upload observations to LoqusDB."""

    def __init__(self, status_api: Store, hk_api: HousekeeperAPI, loqus_api: LoqusdbAPI):
        self.status = status_api
        self.housekeeper = hk_api
        self.loqusdb = loqus_api

    def get_input(self, analysis_obj: models.Analysis) -> ObservationsInputFiles:
        """Fetch input files from an analysis to upload to observations"""

        analysis_date: datetime.datetime = analysis_obj.started_at or analysis_obj.completed_at
        hk_version: models.Version = self.housekeeper.version(
            analysis_obj.family.internal_id, analysis_date
        )

        hk_pedigree: Query = self.housekeeper.files(
            version=hk_version.id, tags=[ObservationAnalysisTag.PEDIGREE]
        ).first()
        hk_snv_gbcf: Query = self.housekeeper.files(
            version=hk_version.id, tags=[ObservationAnalysisTag.CHECK_PROFILE_GBCF]
        ).first()
        hk_snv_vcf: Query = self.housekeeper.files(
            version=hk_version.id, tags=[ObservationAnalysisTag.SNV_VARIANTS]
        ).first()
        sv_vcf_path = None
        if self.loqusdb.analysis_type == ObservationAnalysisType.WGS:
            hk_sv_vcf: Query = self.housekeeper.files(
                version=hk_version.id, tags=[ObservationAnalysisTag.SV_VARIANTS]
            ).first()
            sv_vcf_path: Path = hk_sv_vcf.full_path

        return ObservationsInputFiles(
            case_id=analysis_obj.family.internal_id,
            pedigree=hk_pedigree.full_path,
            snv_gbcf=hk_snv_gbcf.full_path,
            snv_vcf=hk_snv_vcf.full_path,
            sv_vcf=sv_vcf_path,
        )

    def upload(self, input_files: ObservationsInputFiles) -> None:
        """Upload data about genotypes for a family of samples."""

        try:
            existing_case = self.loqusdb.get_case(case_id=input_files.case_id)

        # If CaseNotFoundError is raised, this should trigger the load method of loqusdb
        except CaseNotFoundError:

            duplicate = self.loqusdb.get_duplicate(input_files.snv_gbcf)
            if duplicate:
                err_msg = f"Found duplicate {duplicate['ind_id']} in case {duplicate['case_id']}"
                raise DuplicateSampleError(err_msg)

            results = self.loqusdb.load(
                input_files.case_id,
                input_files.pedigree,
                input_files.snv_vcf,
                input_files.snv_gbcf,
                vcf_sv_path=input_files.sv_vcf,
            )
            LOG.info("parsed %s variants", results["variants"])

        else:
            log_msg = f"found existing family {existing_case['case_id']}, skipping observations"
            LOG.debug(log_msg)

    def process(self, analysis_obj: models.Analysis):
        """Process an upload observation counts for a case."""
        # check if some sample has already been uploaded
        for link in analysis_obj.family.links:
            if link.sample.loqusdb_id:
                raise DuplicateRecordError(f"{link.sample.internal_id} already in LoqusDB")
        results = self.get_input(analysis_obj)
        self.upload(results)
        case_obj = self.loqusdb.get_case(analysis_obj.family.internal_id)
        for link in analysis_obj.family.links:
            link.sample.loqusdb_id = str(case_obj["_id"])
        self.status.commit()

    @staticmethod
    def _all_samples(links: List[models.FamilySample]) -> bool:
        """Return True if all samples are external or sequenced inhouse."""
        return all(
            (link.sample.sequenced_at or link.sample.application_version.application.is_external)
            for link in links
        )



"""Code that handles CLI commands to upload"""
import datetime as dt
import logging

from pathlib import Path

import click

from cg.constants import Pipeline, DataDelivery
from cg.constants.constants import DRY_RUN
from cg.meta.workflow.analysis import AnalysisAPI
from cg.store import Store, models

from cg.apps.tb import TrailblazerAPI
from cg.constants.delivery import PIPELINE_ANALYSIS_TAG_MAP
from cg.constants.priority import PRIORITY_TO_SLURM_QOS
from cg.meta.deliver import DeliverAPI
from cg.meta.rsync import RsyncAPI

LOG = logging.getLogger(__name__)


@click.group()
@click.pass_context
def fastq(context: click.Context):
    context.obj.meta_apis["delivery_api"] = DeliverAPI(
        store=context.obj.status_db,
        hk_api=context.obj.housekeeper_api,
        case_tags=PIPELINE_ANALYSIS_TAG_MAP[DataDelivery.FASTQ]["case_tags"],
        sample_tags=PIPELINE_ANALYSIS_TAG_MAP[DataDelivery.FASTQ]["sample_tags"],
        delivery_type=DataDelivery.FASTQ,
        project_base_path=Path(context.obj.delivery_path),
    )
    context.obj.meta_apis["rsync_api"] = RsyncAPI(context.obj)


@fastq.command("all-available")
@click.pass_context
@DRY_RUN
def auto_fastq(context: click.Context, dry_run: bool):
    """Starts upload of all not previously uploaded cases with analysis type fastq to
    clinical-delivery"""

    status_db: Store = context.obj.status_db
    trailblazer_api: TrailblazerAPI = context.obj.trailblazer_api
    for analysis_obj in status_db.analyses_to_upload(pipeline=Pipeline.FASTQ):
        if analysis_obj.family.analyses[0].uploaded_at:
            LOG.warning(
                "Newer analysis already uploaded for %s, skipping",
                analysis_obj.family.internal_id,
            )
            continue
        if analysis_obj.upload_started_at:
            if trailblazer_api.is_latest_analysis_completed(
                case_id=analysis_obj.family.internal_id
            ):
                LOG.info(
                    "The upload for %s is completed, setting uploaded at to %s",
                    analysis_obj.family.internal_id,
                    dt.datetime.now(),
                )
                analysis_obj.uploaded_at = dt.datetime.now()
                if not dry_run:
                    status_db.commit()
            else:
                LOG.warning(
                    "Upload to clinical-delivery for %s has already started, skipping",
                    analysis_obj.family.internal_id,
                )
            continue
        case: models.Family = analysis_obj.family
        analysis_obj.upload_started_at = dt.datetime.now()
        LOG.info("Uploading family: %s", case.internal_id)
        context.invoke(upload_fastq, case_id=case.internal_id, dry_run=dry_run)
        status_db.commit()


@fastq.command("case")
@click.pass_context
@click.argument("case_id", required=True)
@DRY_RUN
def upload_fastq(context: click.Context, case_id: str, dry_run: bool):
    """Uploads fastq files for a case to clinical-delivery"""

    status_db: Store = context.obj.status_db
    deliver_api: DeliverAPI = context.obj.meta_apis["delivery_api"]
    rsync_api: RsyncAPI = context.obj.meta_apis["rsync_api"]
    trailblazer_api: TrailblazerAPI = context.obj.trailblazer_api

    case: models.Family = status_db.family(internal_id=case_id)
    deliver_api.deliver_files(case)
    job_id: int = rsync_api.slurm_rsync_single_case(
        case_id=case_id, dry_run=dry_run, sample_files_present=True
    )
    rsync_api.write_trailblazer_config(
        {"jobs": [str(job_id)]}, config_path=rsync_api.trailblazer_config_path
    )
    if not dry_run:
        trailblazer_api.add_pending_analysis(
            case_id=case_id,
            analysis_type=AnalysisAPI.get_application_type(
                status_db.family(case_id).links[0].sample
            ),
            config_path=str(rsync_api.trailblazer_config_path),
            out_dir=str(rsync_api.log_dir),
            slurm_quality_of_service=PRIORITY_TO_SLURM_QOS[case.priority],
            data_analysis=Pipeline.FASTQ,
        )
    LOG.info("Transfer of case %s started with SLURM job id %s", case_id, job_id)



"""Test how the api handles versions"""
import datetime


def test_new_version(a_date, housekeeper_api):
    """Test to create a new version with the api"""
    # GIVEN a housekeeper api and a date
    # WHEN creating a version object
    version_obj = housekeeper_api.new_version(created_at=a_date)
    # THEN assert that the version obj was created
    assert version_obj
    # THEN assert that the version object has the correct date
    assert version_obj.created_at == a_date
    # THEN assert that the other members are set to None
    assert version_obj.expires_at is None
    assert version_obj.included_at is None
    assert version_obj.archived_at is None
    assert version_obj.archive_path is None
    # THEN assert that there is no bundle id
    assert version_obj.bundle_id is None


def test_get_version_non_existing(housekeeper_api, a_date):
    """Test to get a version when there are no existing version"""
    # GIVEN a empty housekeeper_api
    # WHEN fetching a version
    version_obj = housekeeper_api.version(bundle=None, date=a_date)
    # THEN assert that the function returns None
    assert version_obj is None


def test_get_version_existing(housekeeper_api, a_date, bundle_data):
    """Test to get a version when there is a bundle and a version"""
    # GIVEN a populated housekeeper_api
    bundle_obj, version_obj = housekeeper_api.add_bundle(bundle_data)
    housekeeper_api.add_commit(bundle_obj, version_obj)
    # WHEN fetching a version
    fetched_version = housekeeper_api.version(bundle=bundle_obj.name, date=a_date)
    # THEN assert that the function returns True and that version has a bundle_id attribute
    assert fetched_version
    assert fetched_version.bundle_id == bundle_obj.id


def test_add_version_existing_bundle(populated_housekeeper_api, later_date, case_id, small_helpers):
    """Test to get a version when there is a bundle and a version"""
    # GIVEN a populated housekeeper_api and a bundle with one version
    bundle_obj = populated_housekeeper_api.bundle(name=case_id)
    assert bundle_obj
    assert small_helpers.length_of_iterable(bundle_obj.versions) == 1
    # WHEN creating a newer version and adding it to the bundle
    new_version = populated_housekeeper_api.new_version(created_at=later_date)
    new_version.bundle = bundle_obj
    populated_housekeeper_api.add_commit(new_version)

    # THEN assert that the bundle has two versions
    bundle_obj = populated_housekeeper_api.bundle(name=case_id)
    assert small_helpers.length_of_iterable(bundle_obj.versions) == 2


def test_get_last_version(populated_housekeeper_api, later_date, case_id):
    """Test to get a version when there is a bundle and a version"""
    # GIVEN a populated housekeeper_api and a bundle with two versions
    bundle_obj = populated_housekeeper_api.bundle(name=case_id)
    new_version = populated_housekeeper_api.new_version(created_at=later_date)
    new_version.bundle = bundle_obj
    populated_housekeeper_api.add_commit(new_version)

    # WHEN fetching the last version of a bundle
    fetched_version = populated_housekeeper_api.last_version(bundle=case_id)

    # THEN assert that the later_date version is fetched
    assert fetched_version.created_at == later_date


def test_get_create_version(populated_housekeeper_api, case_id, bundle_data):

    # Given a populated housekeeper_api with a bundle
    version_obj = populated_housekeeper_api.bundle(case_id).versions[0]

    # When the case_id is given the function should return its latest version
    latest_version = populated_housekeeper_api.get_create_version(case_id)

    # Then assert that the versions match
    assert version_obj == latest_version

    # When a case_id not present in hk is given.
    latest_version = populated_housekeeper_api.get_create_version("new_case")

    # Then assert the new bundle is created the version is new.
    assert latest_version.bundle.name == "new_case"



"""Fixtures for the prepare_fastq api tests"""

import pytest
from tests.meta.compress.conftest import fixture_compress_api, fixture_real_crunchy_api

from cg.meta.compress.compress import CompressAPI
from cg.models.compression_data import CompressionData


@pytest.fixture(scope="function", name="populated_compress_spring_api")
def fixture_populated_compress_spring_api(
    compress_api: CompressAPI, only_spring_bundle: dict, helpers
) -> CompressAPI:
    """Populated compress api fixture with only spring compressed fastq"""
    helpers.ensure_hk_bundle(compress_api.hk_api, only_spring_bundle)

    return compress_api


@pytest.fixture(scope="function", name="populated_compress_api_fastq_spring")
def fixture_populated_compress_api_fastq_spring(
    compress_api: CompressAPI, spring_fastq_mix: dict, helpers
) -> CompressAPI:
    """Populated compress api fixture with both spring and fastq"""
    helpers.ensure_hk_bundle(compress_api.hk_api, spring_fastq_mix)

    return compress_api


@pytest.fixture(name="only_spring_bundle")
def fixture_only_spring_bundle() -> dict:
    """Return a dictionary with bundle info in the correct format"""
    return {
        "name": "ADM1",
        "created": "2019-12-24",
        "files": [
            {
                "path": "/path/HVCHCCCXY-l4t21_535422_S4_L004.spring",
                "archive": False,
                "tags": ["spring"],
            },
        ],
    }


@pytest.fixture(name="spring_fastq_mix")
def fixture_spring_fastq_mix(compression_object: CompressionData) -> dict:
    """Return a dictionary with bundle info including both fastq and spring files"""

    return {
        "name": "ADM1",
        "created": "2019-12-24",
        "files": [
            {
                "path": str(compression_object.spring_path),
                "archive": False,
                "tags": ["spring"],
            },
            {
                "path": str(compression_object.fastq_first),
                "archive": False,
                "tags": ["fastq"],
            },
            {
                "path": str(compression_object.fastq_second),
                "archive": False,
                "tags": ["fastq"],
            },
        ],
    }



import yaml

from cg.meta.workflow.balsamic import BalsamicAnalysisAPI


def get_balsamic_raw_data():
    """Extracts mock BALSAMIC analysis data"""

    config: dict = yaml.safe_load(open("tests/fixtures/apps/balsamic/case/config.json"))
    metrics: dict = yaml.safe_load(
        open("tests/fixtures/apps/balsamic/case/metrics_deliverables.yaml")
    )

    return config, metrics


class MockBalsamicAnalysis(BalsamicAnalysisAPI):
    """Mock a BALSAMIC analysis object"""

    def get_latest_metadata(self, case_id: str):
        """Get the latest metadata of a specific BALSAMIC case"""

        config, metrics = get_balsamic_raw_data()
        return self.parse_analysis(config, metrics)

    def get_bundle_deliverables_type(self, case_id: str) -> str:
        """Return the analysis type for a case"""

        return "tumor_normal_panel"



"""
These are independent interfaces to tools outside the package.

- "Apps" are NOT allowed to cross-talk directly.
- "Apps" are NOT allowed to use functionality of the main package directly.

All access to functions, classes, exceptions from within the main packages should
go through these "app" interfaces.
"""



import logging
from typing import List, Union

from cg.constants import (
    BALSAMIC_REPORT_ACCREDITED_PANELS,
    REQUIRED_REPORT_FIELDS,
    REQUIRED_CUSTOMER_FIELDS,
    REQUIRED_CASE_FIELDS,
    REQUIRED_APPLICATION_FIELDS,
    REQUIRED_DATA_ANALYSIS_BALSAMIC_FIELDS,
    REQUIRED_SAMPLE_BALSAMIC_FIELDS,
    REQUIRED_SAMPLE_METHODS_FIELDS,
    REQUIRED_SAMPLE_TIMESTAMP_FIELDS,
    REQUIRED_SAMPLE_METADATA_BALSAMIC_TARGETED_FIELDS,
    REQUIRED_SAMPLE_METADATA_BALSAMIC_TN_WGS_FIELDS,
    BALSAMIC_ANALYSIS_TYPE,
    REQUIRED_SAMPLE_METADATA_BALSAMIC_TO_WGS_FIELDS,
)
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.meta.report.api import ReportAPI
from cg.models.balsamic.analysis import BalsamicAnalysis
from cg.models.balsamic.metrics import BalsamicTargetedQCMetrics, BalsamicWGSQCMetrics
from cg.models.cg_config import CGConfig
from cg.models.report.metadata import (
    BalsamicTargetedSampleMetadataModel,
    BalsamicWGSSampleMetadataModel,
)
from cg.models.report.report import CaseModel
from cg.models.report.sample import SampleModel
from cg.store import models

LOG = logging.getLogger(__name__)


class BalsamicReportAPI(ReportAPI):
    """API to create BALSAMIC delivery reports"""

    def __init__(self, config: CGConfig, analysis_api: BalsamicAnalysisAPI):
        super().__init__(config=config, analysis_api=analysis_api)
        self.analysis_api = analysis_api

    def get_sample_metadata(
        self, case: models.Family, sample: models.Sample, analysis_metadata: BalsamicAnalysis
    ) -> Union[BalsamicTargetedSampleMetadataModel, BalsamicWGSSampleMetadataModel]:
        """Fetches the sample metadata to include in the report"""

        sample_metrics = analysis_metadata.sample_metrics[sample.internal_id]
        million_read_pairs = round(sample.reads / 2000000, 1) if sample.reads else None

        if "wgs" in self.get_data_analysis_type(case):
            return self.get_wgs_metadata(million_read_pairs, sample_metrics)

        return self.get_panel_metadata(
            sample, million_read_pairs, sample_metrics, analysis_metadata
        )

    @staticmethod
    def get_panel_metadata(
        sample, million_read_pairs, sample_metrics, analysis_metadata
    ) -> BalsamicTargetedSampleMetadataModel:
        """Returns a report metadata for BALSAMIC TGS analysis"""

        return BalsamicTargetedSampleMetadataModel(
            bait_set=sample.capture_kit,
            bait_set_version=analysis_metadata.config.panel.capture_kit_version,
            million_read_pairs=million_read_pairs,
            median_target_coverage=sample_metrics.median_target_coverage
            if sample_metrics
            else None,
            pct_250x=sample_metrics.pct_target_bases_250x if sample_metrics else None,
            pct_500x=sample_metrics.pct_target_bases_500x if sample_metrics else None,
            duplicates=sample_metrics.percent_duplication if sample_metrics else None,
            mean_insert_size=sample_metrics.mean_insert_size if sample_metrics else None,
            fold_80=sample_metrics.fold_80_base_penalty if sample_metrics else None,
        )

    def get_wgs_metadata(
        self, million_read_pairs: float, sample_metrics: BalsamicWGSQCMetrics
    ) -> BalsamicWGSSampleMetadataModel:
        """Returns a report metadata for BALSAMIC WGS analysis"""

        return BalsamicWGSSampleMetadataModel(
            million_read_pairs=million_read_pairs,
            median_coverage=sample_metrics.median_coverage if sample_metrics else None,
            pct_15x=sample_metrics.pct_15x if sample_metrics else None,
            pct_60x=sample_metrics.pct_60x if sample_metrics else None,
            duplicates=self.get_wgs_percent_duplication(sample_metrics),
            mean_insert_size=sample_metrics.mean_insert_size if sample_metrics else None,
            fold_80=sample_metrics.fold_80_base_penalty if sample_metrics else None,
        )

    @staticmethod
    def get_wgs_percent_duplication(sample_metrics: BalsamicWGSQCMetrics):
        """Returns the duplication percentage taking into account both reads"""

        return (
            (sample_metrics.percent_duplication_r1 + sample_metrics.percent_duplication_r2) / 2
            if sample_metrics
            else None
        )

    def get_data_analysis_type(self, case: models.Family) -> str:
        """Retrieves the data analysis type carried out"""

        return self.analysis_api.get_bundle_deliverables_type(case.internal_id)

    def get_genome_build(self, analysis_metadata: BalsamicAnalysis) -> str:
        """Returns the build version of the genome reference of a specific case"""

        return analysis_metadata.config.reference.reference_genome_version

    def get_variant_callers(self, analysis_metadata: BalsamicAnalysis) -> list:
        """Extracts the list of BALSAMIC variant-calling filters from the config.json file"""

        sequencing_type = analysis_metadata.config.analysis.sequencing_type
        analysis_type = analysis_metadata.config.analysis.analysis_type
        var_callers = analysis_metadata.config.vcf

        analysis_var_callers = list()
        for var_caller, var_caller_attributes in var_callers.items():
            if (
                sequencing_type in var_caller_attributes.sequencing_type
                and analysis_type in var_caller_attributes.analysis_type
            ):
                analysis_var_callers.append(var_caller)

        return analysis_var_callers

    def get_report_accreditation(
        self, samples: List[SampleModel], analysis_metadata: BalsamicAnalysis
    ) -> bool:
        """Checks if the report is accredited or not"""

        if analysis_metadata.config.analysis.sequencing_type == "targeted" and next(
            (
                i
                for i in BALSAMIC_REPORT_ACCREDITED_PANELS
                if i in str(analysis_metadata.config.panel.capture_kit)
            ),
            None,
        ):
            return True

        return False

    def get_required_fields(self, case: CaseModel) -> dict:
        """Retrieves a dictionary with the delivery report required fields for BALSAMIC"""

        analysis_type = case.data_analysis.type
        required_sample_metadata_fields = list()

        if BALSAMIC_ANALYSIS_TYPE["tumor_wgs"] in analysis_type:
            required_sample_metadata_fields = REQUIRED_SAMPLE_METADATA_BALSAMIC_TO_WGS_FIELDS
        elif BALSAMIC_ANALYSIS_TYPE["tumor_normal_wgs"] in analysis_type:
            required_sample_metadata_fields = REQUIRED_SAMPLE_METADATA_BALSAMIC_TN_WGS_FIELDS
        elif (
            BALSAMIC_ANALYSIS_TYPE["tumor_panel"] in analysis_type
            or BALSAMIC_ANALYSIS_TYPE["tumor_normal_panel"] in analysis_type
        ):
            required_sample_metadata_fields = REQUIRED_SAMPLE_METADATA_BALSAMIC_TARGETED_FIELDS

        return {
            "report": REQUIRED_REPORT_FIELDS,
            "customer": REQUIRED_CUSTOMER_FIELDS,
            "case": REQUIRED_CASE_FIELDS,
            "applications": self.get_application_required_fields(case, REQUIRED_APPLICATION_FIELDS),
            "data_analysis": REQUIRED_DATA_ANALYSIS_BALSAMIC_FIELDS,
            "samples": self.get_sample_required_fields(case, REQUIRED_SAMPLE_BALSAMIC_FIELDS),
            "methods": self.get_sample_required_fields(case, REQUIRED_SAMPLE_METHODS_FIELDS),
            "timestamps": self.get_sample_required_fields(case, REQUIRED_SAMPLE_TIMESTAMP_FIELDS),
            "metadata": self.get_sample_required_fields(case, required_sample_metadata_fields),
        }



""" This file groups all tests related to microsalt case config creation """

import logging
from pathlib import Path

from cg.apps.lims import LimsAPI
from cg.cli.workflow.microsalt.base import config_case
from cg.models.cg_config import CGConfig
from click.testing import CliRunner
from snapshottest import Snapshot

EXIT_SUCCESS = 0


def test_no_arguments(cli_runner: CliRunner, base_context: CGConfig, caplog):
    """Test command without any options"""

    # GIVEN

    # WHEN dry running without anything specified
    result = cli_runner.invoke(config_case, obj=base_context)

    # THEN command should mention missing arguments
    assert result.exit_code != EXIT_SUCCESS
    assert "Missing argument" in result.output


def test_no_sample_found(cli_runner: CliRunner, base_context: CGConfig, caplog):
    """Test missing sample command"""

    # GIVEN a not existing sample
    microbial_sample_id = "not_existing_sample"

    # WHEN dry running a sample name
    result = cli_runner.invoke(config_case, [microbial_sample_id, "-s"], obj=base_context)

    # THEN command should mention missing sample
    assert result.exit_code != EXIT_SUCCESS
    with caplog.at_level(logging.ERROR):
        assert f"No sample found with id: {microbial_sample_id}" in caplog.text


def test_no_order_found(
    cli_runner: CliRunner, base_context: CGConfig, caplog, invalid_ticket_number: int
):
    """Test missing order command"""

    # GIVEN a not existing ticket
    ticket = invalid_ticket_number

    # WHEN dry running a order name
    result = cli_runner.invoke(config_case, [ticket, "-t"], obj=base_context)

    # THEN command should mention missing order
    assert result.exit_code != EXIT_SUCCESS
    with caplog.at_level(logging.ERROR):
        assert f"No case found for ticket number:  {ticket}" in caplog.text


def test_no_case_found(cli_runner: CliRunner, base_context: CGConfig, caplog):
    """Test missing sample and order command"""

    # GIVEN a not existing order
    microbial_case = "smallzergling"

    # WHEN dry running a order name
    result = cli_runner.invoke(
        config_case,
        [microbial_case],
        obj=base_context,
    )

    # THEN command should mention missing order
    assert result.exit_code != EXIT_SUCCESS
    with caplog.at_level(logging.ERROR):
        assert f"No case found with the id:  {microbial_case}" in caplog.text


def test_dry_sample(
    cli_runner: CliRunner,
    base_context: CGConfig,
    microbial_sample_id: str,
    snapshot: Snapshot,
    lims_api: LimsAPI,
):
    """Test working dry command for sample"""

    # GIVEN project, organism and reference genome is specified in lims
    lims_sample = lims_api.sample(microbial_sample_id)
    lims_sample.sample_data["project"] = {"id": "microbial_order_test"}

    # WHEN dry running a sample name
    result = cli_runner.invoke(config_case, [microbial_sample_id, "-s", "-d"], obj=base_context)

    # THEN command should give us a json dump
    assert result.exit_code == EXIT_SUCCESS


def test_dry_order(
    cli_runner: CliRunner, base_context: CGConfig, microbial_ticket, snapshot: Snapshot
):
    """Test working dry command for a order"""

    # GIVEN

    # WHEN dry running a sample name
    result = cli_runner.invoke(
        config_case,
        [microbial_ticket, "-t", "-d"],
        obj=base_context,
    )

    # THEN command should give us a json dump
    assert result.exit_code == EXIT_SUCCESS


def test_sample(base_context, cli_runner, lims_api, microbial_sample_id, snapshot: Snapshot):
    """Test working command for sample"""

    # GIVEN an existing queries path
    Path(base_context.meta_apis["analysis_api"].queries_path).mkdir(exist_ok=True)
    lims_api.sample(microbial_sample_id).sample_data["project"] = {"id": "microbial_order_test"}

    # WHEN dry running a sample name
    result = cli_runner.invoke(config_case, [microbial_sample_id, "-s"], obj=base_context)

    # THEN command should give us a json dump
    assert result.exit_code == EXIT_SUCCESS


def test_gonorrhoeae(cli_runner: CliRunner, base_context: CGConfig, microbial_sample_id):
    """Test if the substitution of the organism happens"""
    # GIVEN a sample with organism set to gonorrhea
    sample_obj = base_context.meta_apis["analysis_api"].status_db.sample(microbial_sample_id)
    sample_obj.organism.internal_id = "gonorrhoeae"

    # WHEN getting the case config
    result = cli_runner.invoke(config_case, [microbial_sample_id, "-d", "-s"], obj=base_context)

    # THEN the organism should now be  ...
    assert "Neisseria spp." in result.output


def test_cutibacterium_acnes(cli_runner: CliRunner, base_context: CGConfig, microbial_sample_id):
    """Test if this bacteria gets its name changed"""
    # GIVEN a sample with organism set to Cutibacterium acnes
    sample_obj = base_context.meta_apis["analysis_api"].status_db.sample(microbial_sample_id)
    sample_obj.organism.internal_id = "Cutibacterium acnes"

    # WHEN getting the case config
    result = cli_runner.invoke(config_case, [microbial_sample_id, "-s", "-d"], obj=base_context)

    # THEN the organism should now be ....
    assert "Propionibacterium acnes" in result.output


def test_vre_nc_017960(cli_runner: CliRunner, base_context: CGConfig, microbial_sample_id):
    """Test if this bacteria gets its name changed"""
    # GIVEN a sample with organism set to VRE
    sample_obj = base_context.meta_apis["analysis_api"].status_db.sample(microbial_sample_id)
    sample_obj.organism.internal_id = "VRE"
    sample_obj.organism.reference_genome = "NC_017960.1"

    # WHEN getting the case config
    result = cli_runner.invoke(config_case, [microbial_sample_id, "-s", "-d"], obj=base_context)

    # THEN the organism should now be ....
    assert "Enterococcus faecium" in result.output


def test_vre_nc_004668(cli_runner: CliRunner, base_context: CGConfig, microbial_sample_id):
    """Test if this bacteria gets its name changed"""
    # GIVEN a sample with organism set to VRE
    sample_obj = base_context.meta_apis["analysis_api"].status_db.sample(microbial_sample_id)
    sample_obj.organism.internal_id = "VRE"
    sample_obj.organism.reference_genome = "NC_004668.1"

    # WHEN getting the case config
    result = cli_runner.invoke(config_case, [microbial_sample_id, "-s", "-d"], obj=base_context)

    # THEN the organism should now be ....
    assert "Enterococcus faecalis" in result.output


def test_vre_comment(cli_runner: CliRunner, base_context: CGConfig, lims_api, microbial_sample_id):
    """Test if this bacteria gets its name changed"""
    # GIVEN a sample with organism set to VRE and a comment set in LIMS
    sample_obj = base_context.meta_apis["analysis_api"].status_db.sample(microbial_sample_id)
    sample_obj.organism.internal_id = "VRE"
    lims_sample = lims_api.sample(microbial_sample_id)
    lims_sample.sample_data["comment"] = "ABCD123"

    # WHEN getting the case config
    result = cli_runner.invoke(config_case, [microbial_sample_id, "-s", "-d"], obj=base_context)

    # THEN the organism should now be ....
    assert "ABCD123" in result.output



import datetime as dt
import logging

from click.testing import CliRunner
from pathlib import Path

from cg.cli.workflow.commands import rsync_past_run_dirs
from cg.constants.process import EXIT_SUCCESS
from cg.models.cg_config import CGConfig


def test_clean_rsync_past_run_dirs_young_process(
    caplog,
    clean_context: CGConfig,
    cli_runner: CliRunner,
    rsync_process: Path,
    timestamp_today: dt.datetime,
):
    """
    Test that the clean function does not remove newly created dirs.
    rsync_process is always created at the time of the test
    """

    caplog.set_level(logging.INFO)
    # GIVEN a newly created rsync_process
    assert rsync_process.exists()
    # WHEN attempting to remove said process same day
    result = cli_runner.invoke(
        rsync_past_run_dirs, ["-d", "-y", str(timestamp_today)], obj=clean_context
    )
    # THEN the command should not fail and notice that the process is not old
    assert result.exit_code == EXIT_SUCCESS
    assert f"{rsync_process.as_posix()} is still young" in caplog.text


def test_clean_rsync_past_run_dirs_old_process(
    caplog,
    clean_context: CGConfig,
    cli_runner: CliRunner,
    rsync_process: Path,
    timestamp_in_2_weeks: dt.datetime,
):
    """Test that the clean function does remove old dirs. rsync_process is always created at the time of the test"""

    caplog.set_level(logging.INFO)
    # GIVEN a newly created rsync process
    assert rsync_process.exists()
    # WHEN trying to remove it 2 weeks in the future
    result = cli_runner.invoke(
        rsync_past_run_dirs, ["-y", str(timestamp_in_2_weeks)], obj=clean_context
    )
    # THEN it should not fail, inform what process that is removed and it should not exist any more
    assert result.exit_code == EXIT_SUCCESS
    assert f"Removing {rsync_process.as_posix()}" in caplog.text
    assert not rsync_process.exists()



"""Common CLI file generation functions"""

import click


from .mip_dna.base import mip_dna
from .balsamic.base import balsamic


@click.group()
def generate():
    """Generates and/or modifies files"""


generate.add_command(mip_dna)
generate.add_command(balsamic)



"""NIPT ftp upload API"""
import datetime as dt
import logging
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Union

import requests
from requests import Response

import paramiko
from cg.apps.cgstats.stats import StatsAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Pipeline
from cg.exc import HousekeeperFileMissingError, StatinaAPIHTTPError
from cg.meta.upload.nipt.models import StatinaUploadFiles
from cg.models.cg_config import CGConfig
from cg.store import Store, models
from housekeeper.store import models as hk_models

LOG = logging.getLogger(__name__)


class NiptUploadAPI:
    """API for uploading the Fluffy analysis result file to the Klinisk Genetik ftp server"""

    RESULT_FILE_TAGS = ["nipt", "metrics"]

    def __init__(self, config: CGConfig):
        self.sftp_user: str = config.fluffy.sftp.user
        self.sftp_password: str = config.fluffy.sftp.password
        self.sftp_host: str = config.fluffy.sftp.host
        self.statina_user: str = config.statina.user
        self.statina_password: str = config.statina.key
        self.statina_auth_url: str = config.statina.api_url + config.statina.auth_path
        self.statina_upload_url: str = config.statina.api_url + config.statina.upload_path
        self.sftp_port = config.fluffy.sftp.port
        self.sftp_remote_path = config.fluffy.sftp.remote_path
        self.root_dir = Path(config.housekeeper.root)
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.stats_api: StatsAPI = config.cg_stats_api
        self.status_db: Store = config.status_db
        self.dry_run: bool = False

    def set_dry_run(self, dry_run: bool) -> None:
        """Set dry run"""

        LOG.info("Set dry run to %s", dry_run)
        self.dry_run = dry_run

    def target_reads(self, case_id: str) -> float:
        """Return the target reads of the case"""
        case_obj: models.Family = self.status_db.family(case_id)
        application_version = self.status_db.ApplicationVersion.query.filter(
            models.ApplicationVersion.id == case_obj.links[0].sample.application_version_id
        ).first()

        return application_version.application.target_reads

    def flowcell_passed_qc_value(self, case_id: str, q30_threshold: float) -> bool:
        """Check average Q30 and of the latest flowcell related to a case"""
        latest_flow_cell: models.Flowcell = self.status_db.get_latest_flow_cell_on_case(
            family_id=case_id
        )
        flow_cell_reads_and_q30_summary: Dict[
            str, Union[int, float]
        ] = self.stats_api.flow_cell_reads_and_q30_summary(flow_cell_name=latest_flow_cell.name)
        return flow_cell_reads_and_q30_summary[
            "q30"
        ] >= q30_threshold and flow_cell_reads_and_q30_summary["reads"] >= self.target_reads(
            case_id=case_id
        )

    def get_housekeeper_results_file(self, case_id: str, tags: Optional[list] = None) -> str:
        """Get the result file for a NIPT analysis from Housekeeper"""

        if not tags:
            tags: List[str] = self.RESULT_FILE_TAGS

        hk_all_results_file: hk_models.File = self.housekeeper_api.find_file_in_latest_version(
            case_id=case_id, tags=tags
        )

        if not hk_all_results_file:
            raise HousekeeperFileMissingError(
                f"No results files found in Housekeeper for NIPT case {case_id}."
            )

        return hk_all_results_file.path

    def get_results_file_path(self, hk_results_file: str) -> Path:
        """Get the full path to the results file on Hasta"""

        results_file: Path = self.root_dir / hk_results_file

        if not results_file.exists():
            raise FileNotFoundError(f"Results file {results_file} not found on hasta!")

        return results_file

    def get_all_upload_analyses(self) -> Iterable[models.Analysis]:
        """Gets all nipt analyses that are ready to be uploaded"""

        latest_nipt_analyses = self.status_db.latest_analyses().filter(
            models.Analysis.pipeline == Pipeline.FLUFFY
        )

        return latest_nipt_analyses.filter(models.Analysis.uploaded_at.is_(None))

    def upload_to_ftp_server(self, results_file: Path) -> None:
        """Upload the result file to the ftp server"""
        if self.dry_run:
            LOG.info(f"Would upload results file to ftp server: {results_file}")
            return
        transport: paramiko.Transport = paramiko.Transport((self.sftp_host, self.sftp_port))
        LOG.info(f"Connecting to SFTP server {self.sftp_host}")
        transport.connect(username=self.sftp_user, password=self.sftp_password)
        sftp: paramiko.SFTPClient = paramiko.SFTPClient.from_transport(transport)
        LOG.info(
            f"Uploading file {results_file} to remote path {self.sftp_remote_path}/{results_file.name}"
        )

        sftp.put(
            localpath=str(results_file),
            remotepath=f"/{self.sftp_remote_path}/{results_file.name}",
            confirm=False,
        )
        sftp.close()
        transport.close()

    def update_analysis_uploaded_at_date(self, case_id: str) -> models.Analysis:
        """Updates analysis_uploaded_at for the uploaded analysis"""

        case_obj: models.Family = self.status_db.family(case_id)
        analysis_obj: models.Analysis = case_obj.analyses[0]

        if not self.dry_run:
            analysis_obj.uploaded_at = dt.datetime.now()
            self.status_db.commit()

        return analysis_obj

    def update_analysis_upload_started_date(self, case_id: str) -> models.Analysis:
        """Updates analysis_upload_started_at for the uploaded analysis"""

        case_obj: models.Family = self.status_db.family(case_id)
        analysis_obj: models.Analysis = case_obj.analyses[0]

        if not self.dry_run:
            analysis_obj.upload_started_at = dt.datetime.now()
            self.status_db.commit()

        return analysis_obj

    def get_statina_files(self, case_id: str) -> StatinaUploadFiles:
        """Get statina files from from housekeeper."""

        hk_results_file: str = self.get_housekeeper_results_file(
            case_id=case_id, tags=["nipt", "metrics"]
        )
        results_file: Path = self.get_results_file_path(hk_results_file)

        hk_multiqc_file: str = self.get_housekeeper_results_file(
            case_id=case_id, tags=["nipt", "multiqc-html"]
        )
        multiqc_file: Path = self.get_results_file_path(hk_multiqc_file)

        hk_segmental_calls_file: str = self.get_housekeeper_results_file(
            case_id=case_id, tags=["nipt", "wisecondor"]
        )
        segmental_calls_file: Path = self.get_results_file_path(hk_segmental_calls_file)

        return StatinaUploadFiles(
            result_file=results_file.absolute().as_posix(),
            multiqc_report=multiqc_file.absolute().as_posix(),
            segmental_calls=segmental_calls_file.parent.as_posix(),
        )

    def upload_to_statina_database(self, statina_files: StatinaUploadFiles):
        """Upload nipt data via rest-API."""

        token: str = (
            requests.post(
                self.statina_auth_url,
                data={"username": self.statina_user, "password": self.statina_password},
            )
            .json()
            .get("access_token")
        )

        response: Response = requests.post(
            url=self.statina_upload_url,
            headers={"authorization": f"Bearer {token}"},
            data=statina_files.json(exclude_none=True),
        )
        if not response.ok:
            LOG.error(response.text)
            raise StatinaAPIHTTPError(response.text)

        LOG.info("nipt output: %s", response.text)



"""Utility functions for the upload cli commands"""
import logging
from typing import List, Optional

import click
from cg.constants import Pipeline
from cg.store import Store, models

LOG = logging.getLogger(__name__)


class LinkHelper:
    """Class that helps handle links"""

    @staticmethod
    def all_samples_are_non_tumour(links: List[models.FamilySample]) -> bool:
        """Return True if all samples are non tumour."""
        return all(not link.sample.is_tumour for link in links)

    @staticmethod
    def get_analysis_type_for_each_link(links: List[models.FamilySample]) -> list:
        """Return analysis type for each sample given by link list"""
        return [link.sample.application_version.application.analysis_type for link in links]


def suggest_cases_to_upload(status_db: Store, pipeline: Optional[Pipeline] = None) -> None:
    LOG.warning("Provide a case, suggestions:")
    records = status_db.analyses_to_upload(pipeline=pipeline)[:50]
    for case_obj in records:
        click.echo(case_obj)



import logging
from pathlib import Path
from typing import Dict

from pydantic import BaseModel, validator

from cg.apps.cgstats.dragen_demux_sample import DragenDemuxSample
from cg.apps.cgstats.parsers.conversion_stats import ConversionStats, SampleConversionResults
from cg.apps.cgstats.parsers.demux_stats import DemuxStats, SampleBarcodeStats
from cg.models.demultiplex.demux_results import DemuxResults
from cgmodels.demultiplex.sample_sheet import NovaSeqSample, SampleSheet

LOG = logging.getLogger(__name__)


class DemuxSample(BaseModel):
    """Gather statistics from the demultiplexing results for a sample"""

    sample_name: str
    flowcell: str
    lane: int
    nr_raw_clusters_: int
    barcode_stats_: SampleBarcodeStats
    conversion_stats_: SampleConversionResults

    # derived fields etc
    raw_clusters_pc: float = 0.0
    pass_filter_yield_pc: float = 0.0
    pass_filter_clusters: int = 0
    pass_filter_yield: int = 0

    pass_filter_Q30: float = 0
    pass_filter_read1_q30: float = 0
    pass_filter_read2_q30: float = 0
    pass_filter_qscore: float = 0
    undetermined_pc: float = 0
    barcodes: int = 0
    perfect_barcodes: int = 0
    one_mismatch_barcodes: int = 0

    @validator("raw_clusters_pc", always=True)
    def set_raw_clusters_pc(cls, value: float, values: dict) -> float:
        """Calculate the percentage of raw clusters"""
        conversion_stats: SampleConversionResults = values["conversion_stats_"]
        if not values["nr_raw_clusters_"]:
            return value
        return round(conversion_stats.raw_cluster_count / values["nr_raw_clusters_"] * 100, 2)

    @validator("pass_filter_yield_pc", always=True)
    def set_pass_filter_yield_pc(cls, value: float, values: dict) -> float:
        """Calculate the pass filter yield percentage"""
        conversion_stats: SampleConversionResults = values["conversion_stats_"]
        if not values["conversion_stats_"].raw_yield:
            return value
        return round(
            conversion_stats.pass_filter_yield / conversion_stats.raw_yield * 100,
            2,
        )

    @validator("pass_filter_clusters", always=True)
    def set_pass_filter_clusters(cls, _, values) -> int:
        """Set the number of pass filter clusters"""
        conversion_stats: SampleConversionResults = values["conversion_stats_"]
        return conversion_stats.pass_filter_cluster_count

    @validator("pass_filter_yield", always=True)
    def set_pass_filter_yield(cls, _, values) -> int:
        """Set the number of pass filter yield (reads)"""
        conversion_stats: SampleConversionResults = values["conversion_stats_"]
        return conversion_stats.pass_filter_yield

    @validator("pass_filter_Q30", always=True)
    def set_pass_filter_Q30(cls, value, values) -> float:
        """Set the percentage of pass filter high quality reads"""
        conversion_stats: SampleConversionResults = values["conversion_stats_"]
        if not conversion_stats.pass_filter_yield:
            return value
        return round(conversion_stats.pass_filter_q30 / conversion_stats.pass_filter_yield * 100, 2)

    @validator("pass_filter_read1_q30", always=True)
    def set_percentage_high_quality_read_one(cls, value, values) -> float:
        """Calculate the percentage of high quality read one reads of all reads"""
        conversion_stats: SampleConversionResults = values["conversion_stats_"]
        if not conversion_stats.pass_filter_read1_yield:
            return value
        return round(
            conversion_stats.pass_filter_read1_q30 / conversion_stats.pass_filter_read1_yield * 100,
            2,
        )

    @validator("pass_filter_read2_q30", always=True)
    def set_percentage_high_quality_read_two(cls, value, values) -> float:
        """Calculate the percentage of high quality read two reads of all reads"""
        conversion_stats: SampleConversionResults = values["conversion_stats_"]
        if not conversion_stats.pass_filter_read2_yield:
            return value
        return round(
            conversion_stats.pass_filter_read2_q30 / conversion_stats.pass_filter_read2_yield * 100,
            2,
        )

    @validator("pass_filter_qscore", always=True)
    def set_pass_filter_qscore(cls, value, values) -> float:
        """Calculate the average quality score per read"""
        conversion_stats: SampleConversionResults = values["conversion_stats_"]
        if not conversion_stats.pass_filter_yield:
            return value
        return round(
            conversion_stats.pass_filter_quality_score_sum / conversion_stats.pass_filter_yield,
            2,
        )

    @validator("undetermined_pc", always=True)
    def set_undetermined_pc(cls, value, values) -> float:
        """Calculate the average quality score per read"""
        conversion_stats: SampleConversionResults = values["conversion_stats_"]
        barcode_stats: SampleBarcodeStats = values["barcode_stats_"]
        if not conversion_stats.pass_filter_cluster_count:
            return value
        return round(
            (conversion_stats.pass_filter_cluster_count - barcode_stats.barcode_count)
            / conversion_stats.pass_filter_cluster_count
            * 100,
            2,
        )

    @validator("barcodes", always=True)
    def set_nr_barcodes(cls, _, values) -> int:
        """Set the total number of barcodes"""
        barcode_stats: SampleBarcodeStats = values["barcode_stats_"]
        return barcode_stats.barcode_count or 0

    @validator("perfect_barcodes", always=True)
    def set_nr_perfect_barcodes(cls, _, values) -> int:
        """Set the total number of perfect barcodes"""
        barcode_stats: SampleBarcodeStats = values["barcode_stats_"]
        return barcode_stats.perfect_barcode_count or 0

    @validator("one_mismatch_barcodes", always=True)
    def set_nr_one_mismatch_barcodes(cls, _, values) -> int:
        """Set the total number of mismatch barcodes"""
        barcode_stats: SampleBarcodeStats = values["barcode_stats_"]
        return barcode_stats.one_mismatch_barcode_count or 0


def get_dragen_demux_samples(
    demux_results: DemuxResults,
    sample_sheet: SampleSheet,
) -> Dict[int, Dict[str, DragenDemuxSample]]:
    """Gather information from Dragen demultiplexing results and create samples with the correct
    information"""
    LOG.info("Gather post Dragen demultiplexing statistics for demultiplexed samples")
    demux_samples: Dict[int, Dict[str, DragenDemuxSample]] = {}

    for sample in sample_sheet.samples:
        if sample.lane not in demux_samples:
            demux_samples[sample.lane] = {}

        demultiplexing_stats = demux_results.demultiplexing_stats.parsed_stats[sample.lane][
            sample.sample_id
        ]
        adapter_metrics = demux_results.adapter_metrics.parsed_metrics[sample.lane][
            sample.sample_id
        ]

        demux_samples[sample.lane][sample.sample_id] = DragenDemuxSample(
            sample_name=sample.sample_id,
            flowcell=demux_results.flowcell.flowcell_id,
            lane=sample.lane,
            reads=int(demultiplexing_stats["# Reads"]),
            perfect_reads=int(demultiplexing_stats["# Perfect Index Reads"]),
            one_mismatch_reads=int(demultiplexing_stats["# One Mismatch Index Reads"]),
            pass_filter_q30=int(demultiplexing_stats["# of >= Q30 Bases (PF)"]),
            mean_quality_score=float(demultiplexing_stats["Mean Quality Score (PF)"]),
            r1_sample_bases=int(adapter_metrics["R1_SampleBases"]),
            r2_sample_bases=int(adapter_metrics["R2_SampleBases"]),
            read_length=demux_results.run_info.mean_read_length(),
        )

    return demux_samples


def get_demux_samples(
    conversion_stats: ConversionStats, demux_stats_path: Path, sample_sheet: SampleSheet
) -> Dict[int, Dict[str, DemuxSample]]:
    """Gather information from demultiplexing results and create samples with the correct
    information"""
    LOG.info("Gather post demultiplexing statistics for demultiplexed samples")
    demux_samples: Dict[int, Dict[str, DemuxSample]] = {}
    demux_stats: DemuxStats = DemuxStats(demux_stats_path=demux_stats_path)
    raw_clusters: Dict[int, int] = conversion_stats.raw_clusters_per_lane
    flowcell_id: str = conversion_stats.flowcell_id
    sample: NovaSeqSample
    for sample in sample_sheet.samples:
        lane: int = sample.lane
        if lane not in demux_samples:
            demux_samples[lane] = {}
        barcode = (
            "+".join([sample.index, sample.second_index]) if sample.second_index else sample.index
        )
        sample_barcode_stats: SampleBarcodeStats = demux_stats.lanes_to_barcode[lane][barcode]
        sample_conversion_stats: SampleConversionResults = conversion_stats.lanes_to_barcode[lane][
            barcode
        ]
        lane_raw_cluster: int = raw_clusters[lane]
        sample_id: str = sample.sample_id
        demux_samples[lane][sample_id] = DemuxSample(
            sample_name=sample.sample_id,
            flowcell=flowcell_id,
            lane=lane,
            nr_raw_clusters_=lane_raw_cluster,
            barcode_stats_=sample_barcode_stats,
            conversion_stats_=sample_conversion_stats,
        )

    return demux_samples



from enum import Enum, IntEnum


class Gender(str, Enum):
    FEMALE = "female"
    MALE = "male"
    UNKNOWN = "unknown"
    OTHER = "other"
    MISSING = None


class PhenotypeStatus(str, Enum):
    UNKNOWN = "unknown"
    UNAFFECTED = "unaffected"
    AFFECTED = "affected"
    MISSING = None


class PlinkPhenotypeStatus(IntEnum):
    UNKNOWN = 0
    UNAFFECTED = 1
    AFFECTED = 2


class PlinkGender(str, Enum):
    UNKNOWN = 0
    MALE = 1
    FEMALE = 2


class RelationshipStatus(str, Enum):
    HAS_NO_PARENT = 0



from click.testing import CliRunner

from cg.cli.workflow.commands import (
    fluffy_past_run_dirs,
    balsamic_past_run_dirs,
    mip_past_run_dirs,
    mutant_past_run_dirs,
)
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig


def test_cli_workflow_clean_balsamic(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):

    # GIVEN a before string

    # WHEN running command in dry-run
    result = cli_runner.invoke(balsamic_past_run_dirs, [before_date], obj=base_context)

    # THEN command should terminate successfully
    assert result.exit_code == EXIT_SUCCESS


def test_cli_workflow_clean_fluffy(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):

    # GIVEN a before string

    # WHEN running command in dry-run
    result = cli_runner.invoke(fluffy_past_run_dirs, [before_date], obj=base_context)

    # THEN command should terminate successfully
    assert result.exit_code == EXIT_SUCCESS


def test_cli_workflow_clean_mip(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):

    # GIVEN a before string

    # WHEN running command in dry-run
    result = cli_runner.invoke(mip_past_run_dirs, [before_date], obj=base_context)

    # THEN command should terminate successfully
    assert result.exit_code == EXIT_SUCCESS


def test_cli_workflow_clean_mutant(
    cli_runner: CliRunner,
    base_context: CGConfig,
    before_date: str,
):

    # GIVEN a before string

    # WHEN running command in dry-run
    result = cli_runner.invoke(mutant_past_run_dirs, [before_date], obj=base_context)

    # THEN command should terminate successfully
    assert result.exit_code == EXIT_SUCCESS



from pathlib import Path
import pytest
import yaml


@pytest.fixture(name="balsamic_config_path")
def fixture_balsamic_config_path(fixtures_dir) -> Path:
    """Returns path to BALSAMIC case_config.json"""

    return Path(fixtures_dir, "apps", "balsamic", "case", "config.json")


@pytest.fixture(name="balsamic_metrics_path")
def fixture_balsamic_metrics_path(fixtures_dir) -> Path:
    """Returns path to BALSAMIC case_metrics_deliverables.yaml"""

    return Path(fixtures_dir, "apps", "balsamic", "case", "metrics_deliverables.yaml")


@pytest.fixture(name="balsamic_config_raw")
def fixture_balsamic_config(balsamic_config_path) -> dict:
    """Return BALSAMIC config file as a dictionary"""

    with open(Path(balsamic_config_path), "r") as stream:
        config = yaml.safe_load(stream)

    return config


@pytest.fixture(name="balsamic_metrics_raw")
def fixture_balsamic_metrics(balsamic_metrics_path) -> dict:
    """Return BALSAMIC metrics file as a dictionary"""

    with open(Path(balsamic_metrics_path), "r") as stream:
        metrics = yaml.safe_load(stream)

    return metrics



import logging
from typing import Iterable, List

import click
from cg.cli.compress.helpers import get_fastq_individuals, update_compress_api
from cg.exc import CaseNotFoundError
from cg.meta.compress import CompressAPI
from cg.models.cg_config import CGConfig
from cg.store import Store, models

LOG = logging.getLogger(__name__)


@click.command("sample")
@click.argument("sample-id", type=str)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_obj
def store_sample(context: CGConfig, sample_id: str, dry_run: bool):
    """Include links to decompressed FASTQ files belonging to this sample in housekeeper"""
    compress_api: CompressAPI = context.meta_apis["compress_api"]
    status_db: Store = context.status_db
    update_compress_api(compress_api, dry_run=dry_run)

    sample_obj = status_db.sample(internal_id=sample_id)
    was_decompressed: bool = compress_api.add_decompressed_fastq(sample_obj)
    if not was_decompressed:
        LOG.info(f"Skipping sample {sample_id}")
        return 0
    LOG.info(f"Stored fastq files for {sample_id}")
    return 1


@click.command("case")
@click.argument("case-id", type=str)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def store_case(context: click.Context, case_id: str, dry_run: bool):
    """Include links to decompressed FASTQ files belonging to this case in housekeeper"""

    status_db: Store = context.obj.status_db
    try:
        samples: Iterable[str] = get_fastq_individuals(status_db, case_id)
        stored_individuals = 0
        for sample_id in samples:
            stored_count: int = context.invoke(store_sample, sample_id=sample_id, dry_run=dry_run)
            stored_individuals += stored_count
    except CaseNotFoundError:
        return
    LOG.info(f"Stored fastq files for {stored_individuals} samples")


@click.command("flowcell")
@click.argument("flowcell_id", type=str)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def store_flowcell(context: click.Context, flowcell_id: str, dry_run: bool):
    """Include links to decompressed FASTQ files belonging to this flowcell in housekeeper"""

    status_db: Store = context.obj.status_db
    samples: List[models.Sample] = status_db.get_samples_from_flowcell(flowcell_name=flowcell_id)
    stored_individuals = 0
    for sample in samples:
        stored_count: int = context.invoke(
            store_sample, sample_id=sample.internal_id, dry_run=dry_run
        )
        stored_individuals += stored_count
    LOG.info(f"Stored fastq files for {stored_individuals} samples")


@click.command("ticket")
@click.argument("ticket_id", type=int)
@click.option("-d", "--dry-run", is_flag=True)
@click.pass_context
def store_ticket(context: click.Context, ticket_id: int, dry_run: bool):
    """Include links to decompressed FASTQ files belonging to this ticket in housekeeper"""
    status_db: Store = context.obj.status_db
    samples: List[models.Sample] = status_db.get_samples_from_ticket(ticket_id=ticket_id)
    stored_individuals = 0
    for sample in samples:
        stored_count: int = context.invoke(
            store_sample, sample_id=sample.internal_id, dry_run=dry_run
        )
        stored_individuals += stored_count
    LOG.info(f"Stored fastq files for {stored_individuals} samples")



"""Fixtures for lims tests"""

import pytest

from cg.apps.lims.api import LimsAPI


class MockLims(LimsAPI):
    """Mock parts of the lims api"""

    lims = None

    def __init__(self):
        """Mock the init method"""
        self.lims = self

    def get_prepmethod(self, lims_id: str) -> str:
        """Override the get_prepmethod"""

    def get_sequencingmethod(self, lims_id: str) -> str:
        """Override the get_sequencingmethod"""

    def get_deliverymethod(self, lims_id: str) -> str:
        """Override the get_deliverymethod"""


@pytest.fixture(scope="function")
def lims_api():
    """Returns a Lims api mock"""
    _lims_api = MockLims()
    return _lims_api


@pytest.fixture
def skeleton_orderform_sample():
    """Get the skeleton for a orderform"""
    return {
        "UDF/priority": "",
        "UDF/Sequencing Analysis": "",
        "UDF/customer": "",
        "Sample/Name": "",
    }



"""This script tests the cli methods to set values of cases in status-db"""
from click.testing import CliRunner

from cg.cli.set.family import family
from cg.constants import DataDelivery, Pipeline
from cg.models.cg_config import CGConfig
from cg.store import Store

SUCCESS = 0


def test_set_family_without_options(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test to set a case using only the required arguments"""
    # GIVEN a database with a case
    case_id = helpers.add_case(base_store).internal_id
    assert base_store.Family.query.count() == 1

    # WHEN setting a case
    result = cli_runner.invoke(family, [case_id], obj=base_context)

    # THEN it should abort
    assert result.exit_code != SUCCESS


def test_set_family_bad_family(cli_runner, base_context):
    """Test to set a case using a non-existing case"""
    # GIVEN an empty database

    # WHEN setting a case
    case_id = "dummy_name"
    result = cli_runner.invoke(family, [case_id], obj=base_context)

    # THEN it should complain on missing case
    assert result.exit_code != SUCCESS


def test_set_family_bad_panel(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test to set a case using a non-existing panel"""
    # GIVEN a database with a case

    # WHEN setting a case
    panel_id = "dummy_panel"
    case_id = helpers.add_case(base_store).internal_id
    result = cli_runner.invoke(family, [case_id, "--panel", panel_id], obj=base_context)

    # THEN it should complain in missing panel instead of setting a value
    assert result.exit_code != SUCCESS
    assert panel_id not in base_store.Family.query.first().panels


def test_set_family_panel(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test to set a case using an existing panel"""
    # GIVEN a database with a case and a panel not yet added to the case
    panel_id = helpers.ensure_panel(base_store, "a_panel").name
    case_id = helpers.add_case(base_store).internal_id
    assert panel_id not in base_store.Family.query.first().panels

    # WHEN setting a panel of a case
    result = cli_runner.invoke(family, [case_id, "--panel", panel_id], obj=base_context)

    # THEN it should set panel on the case
    assert result.exit_code == SUCCESS
    assert panel_id in base_store.Family.query.first().panels


def test_set_family_priority(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test that the added case gets the priority we send in"""
    # GIVEN a database with a case
    case_id = helpers.add_case(base_store).internal_id
    priority = "priority"
    assert base_store.Family.query.first().priority_human != priority

    # WHEN setting a case
    result = cli_runner.invoke(
        family, [case_id, "--priority", priority], obj=base_context, catch_exceptions=False
    )

    # THEN it should have been set
    assert result.exit_code == SUCCESS
    assert base_store.Family.query.count() == 1
    assert base_store.Family.query.first().priority_human == priority


def test_set_family_customer(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test to set a case using an existing customer"""
    # GIVEN a database with a case and a customer not yet on the case
    customer_id = helpers.ensure_customer(base_store, customer_id="a_customer").internal_id
    case = helpers.add_case(base_store)
    assert customer_id != case.customer.internal_id

    # WHEN setting a customer of a case
    result = cli_runner.invoke(
        family, [case.internal_id, "--customer-id", customer_id], obj=base_context
    )

    # THEN it should set customer on the case
    assert result.exit_code == SUCCESS
    assert customer_id == case.customer.internal_id


def test_set_family_bad_data_analysis(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test to set a case using a non-existing data_analysis"""
    # GIVEN a database with a case

    # WHEN setting a data_analysis on a case
    data_analysis = "dummy_pipeline"
    case_id = helpers.add_case(base_store).internal_id
    result = cli_runner.invoke(
        family, [case_id, "--data-analysis", data_analysis], obj=base_context
    )

    # THEN it should complain in non valid data_analysis instead of setting a value
    assert result.exit_code != SUCCESS
    assert str(data_analysis) != base_store.Family.query.first().data_analysis


def test_set_family_data_analysis(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test to set a case using an existing data_analysis"""

    # GIVEN a database with a case and a data_analysis not yet set on the case
    data_analysis = Pipeline.FASTQ
    case_obj = helpers.add_case(base_store)
    assert str(data_analysis) != case_obj.data_analysis

    # WHEN setting a data_analysis of a case
    result = cli_runner.invoke(
        family,
        [case_obj.internal_id, "--data-analysis", str(data_analysis)],
        obj=base_context,
    )

    # THEN it should set data_analysis on the case
    assert result.exit_code == SUCCESS
    assert str(data_analysis) == case_obj.data_analysis


def test_set_family_bad_data_delivery(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test to set a case using a non-existing data_delivery"""
    # GIVEN a database with a case

    # WHEN setting a data_delivery on a case
    data_delivery = "dummy_delivery"
    case_id = helpers.add_case(base_store).internal_id
    result = cli_runner.invoke(
        family, [case_id, "--data-delivery", data_delivery], obj=base_context
    )

    # THEN it should complain in non valid data_delivery instead of setting a value
    assert result.exit_code != SUCCESS
    assert str(data_delivery) != base_store.Family.query.first().data_delivery


def test_set_family_data_delivery(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    """Test to set a case using an existing data_delivery"""

    # GIVEN a database with a case and a data_delivery not yet set on the case
    data_delivery = DataDelivery.FASTQ
    case_obj = helpers.add_case(base_store)
    assert str(data_delivery) != case_obj.data_delivery

    # WHEN setting a data_delivery of a case
    result = cli_runner.invoke(
        family,
        [case_obj.internal_id, "--data-delivery", str(data_delivery)],
        obj=base_context,
    )

    # THEN it should set data_delivery on the case
    assert result.exit_code == SUCCESS
    assert str(data_delivery) == case_obj.data_delivery



"""Parse statistics from the dragen adapter metrics file"""
import csv
import logging
from pathlib import Path
from typing import Dict

LOG = logging.getLogger(__name__)


class AdapterMetrics:
    def __init__(self, adapter_metrics_path: Path):
        self.adapter_metrics_path = adapter_metrics_path
        self.parsed_metrics = self.parse_metrics_file()

    def parse_metrics_file(
        self,
    ) -> Dict[int, dict]:
        """Parse the Dragen adapter metrics file"""
        LOG.info("Parsing Dragen demultiplexing adapter metrics file %s", self.adapter_metrics_path)

        parsed_metrics = {}

        with open(self.adapter_metrics_path, mode="r") as metrics_file:
            metrics_reader = csv.DictReader(metrics_file)
            for row in metrics_reader:
                lane = int(row["Lane"])
                sample_id = row["Sample_ID"]
                parsed_metrics[lane] = parsed_metrics.get(lane, {})
                parsed_metrics[lane][sample_id] = row

        return parsed_metrics



"""Commands to generates and/or modify MIP rare disease DNA files"""

import click

from cg.meta.report.mip_dna import MipDNAReportAPI
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.cli.generate.commands import delivery_report, available_delivery_reports


@click.group("mip-dna", invoke_without_command=True)
@click.pass_context
def mip_dna(context: click.Context):
    """Rare disease DNA file generation/modification"""

    if context.invoked_subcommand is None:
        click.echo(context.get_help())
        return

    context.obj.meta_apis["report_api"] = MipDNAReportAPI(
        config=context.obj, analysis_api=MipDNAAnalysisAPI(config=context.obj)
    )


mip_dna.add_command(delivery_report)
mip_dna.add_command(available_delivery_reports)



from cg.constants import ANALYSIS_SOURCES, METAGENOME_SOURCES
from cg.models.orders.order import OrderType
from cg.utils.enums import StrEnum

SEX_MAP = {"male": "M", "female": "F", "unknown": "unknown"}
REV_SEX_MAP = {value: key for key, value in SEX_MAP.items()}
CONTAINER_TYPES = ["Tube", "96 well plate"]
SOURCE_TYPES = set().union(METAGENOME_SOURCES, ANALYSIS_SOURCES)

CASE_PROJECT_TYPES = [
    str(OrderType.MIP_DNA),
    str(OrderType.BALSAMIC),
    str(OrderType.MIP_RNA),
]


class Orderform(StrEnum):
    BALSAMIC: str = "1508"
    BALSAMIC_UMI: str = "1508"
    FASTQ: str = "1508"
    FLUFFY: str = "1604"
    METAGENOME: str = "1605"
    MICROSALT: str = "1603"
    MIP_DNA: str = "1508"
    MIP_RNA: str = "1508"
    RML: str = "1604"
    SARS_COV_2: str = "2184"



"""Module for Balsamic Analysis API"""

import json
import logging
from pathlib import Path
from typing import List, Optional, Union, Any

import yaml
from pydantic import ValidationError
from cg.constants import DataDelivery, Pipeline
from cg.constants.tags import BalsamicAnalysisTag
from cg.exc import BalsamicStartError, CgError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.fastq import BalsamicFastqHandler
from cg.models.balsamic.analysis import BalsamicAnalysis
from cg.models.balsamic.metrics import (
    BalsamicTargetedQCMetrics,
    BalsamicWGSQCMetrics,
    BalsamicMetricsBase,
)
from cg.models.cg_config import CGConfig
from cg.store import models
from cg.utils import Process

LOG = logging.getLogger(__name__)


class BalsamicAnalysisAPI(AnalysisAPI):
    """Handles communication between BALSAMIC processes
    and the rest of CG infrastructure"""

    __BALSAMIC_APPLICATIONS = {"wgs", "wes", "tgs"}
    __BALSAMIC_BED_APPLICATIONS = {"wes", "tgs"}

    def __init__(
        self,
        config: CGConfig,
        pipeline: Pipeline = Pipeline.BALSAMIC,
    ):
        super().__init__(config=config, pipeline=pipeline)
        self.root_dir = config.balsamic.root
        self.account = config.balsamic.slurm.account
        self.balsamic_cache = config.balsamic.balsamic_cache
        self.email = config.balsamic.slurm.mail_user
        self.qos = config.balsamic.slurm.qos
        self.bed_path = config.bed_path

    @property
    def root(self) -> str:
        return self.root_dir

    @property
    def threshold_reads(self):
        return True

    @property
    def fastq_handler(self):
        return BalsamicFastqHandler

    @property
    def process(self):
        if not self._process:
            self._process = Process(self.config.balsamic.binary_path)
        return self._process

    def get_case_path(self, case_id: str) -> Path:
        """Returns a path where the Balsamic case for the case_id should be located"""
        return Path(self.root_dir, case_id)

    def get_cases_to_analyze(self) -> List[models.Family]:
        cases_query: List[models.Family] = self.status_db.cases_to_analyze(
            pipeline=self.pipeline, threshold=self.threshold_reads
        )
        cases_to_analyze = []
        for case_obj in cases_query:
            if case_obj.action == "analyze" or not case_obj.latest_analyzed:
                cases_to_analyze.append(case_obj)
            elif (
                self.trailblazer_api.get_latest_analysis_status(case_id=case_obj.internal_id)
                == "failed"
            ):
                cases_to_analyze.append(case_obj)
        return cases_to_analyze

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """Returns a path where the Balsamic deliverables file for the case_id should be located.

        (Optional) Checks if deliverables file exists
        """
        return Path(
            self.root_dir,
            case_id,
            "analysis",
            "delivery_report",
            case_id + ".hk",
        )

    def get_case_config_path(self, case_id: str) -> Path:
        """Generates a path where the Balsamic config for the case_id should be located.

        (Optional) Checks if config file exists.
        """
        return Path(self.root_dir, case_id, case_id + ".json")

    def get_trailblazer_config_path(self, case_id: str) -> Path:
        return Path(self.root_dir, case_id, "analysis", "slurm_jobids.yaml")

    def get_bundle_deliverables_type(self, case_id: str) -> str:
        """Return the analysis type for a case

        Analysis types are any of ["tumor_wgs", "tumor_normal_wgs", "tumor_panel", "tumor_normal_panel"]
        """
        LOG.debug("Fetch analysis type for %s", case_id)
        number_of_samples: int = len(self.status_db.family(case_id).links)

        application_type: str = self.get_application_type(
            self.status_db.family(case_id).links[0].sample
        )
        sample_type = "tumor"
        if number_of_samples == 2:
            sample_type = "tumor_normal"
        if application_type != "wgs":
            application_type = "panel"
        analysis_type = "_".join([sample_type, application_type])
        LOG.info("Found analysis type %s", analysis_type)
        return analysis_type

    def get_sample_fastq_destination_dir(
        self, case_obj: models.Family, sample_obj: models.Sample
    ) -> Path:
        return self.get_case_path(case_obj.internal_id) / "fastq"

    def link_fastq_files(self, case_id: str, dry_run: bool = False) -> None:
        case_obj = self.status_db.family(case_id)
        for link in case_obj.links:
            self.link_fastq_files_for_sample(
                case_obj=case_obj, sample_obj=link.sample, concatenate=True
            )

    def get_concatenated_fastq_path(self, link_object: models.FamilySample) -> Path:
        """Returns path to the concatenated FASTQ file of a sample"""
        file_collection: List[dict] = self.gather_file_metadata_for_sample(link_object.sample)
        fastq_data = file_collection[0]
        linked_fastq_name = self.fastq_handler.create_fastq_name(
            lane=fastq_data["lane"],
            flowcell=fastq_data["flowcell"],
            sample=link_object.sample.internal_id,
            read=fastq_data["read"],
            undetermined=fastq_data["undetermined"],
        )
        concatenated_fastq_name: str = self.fastq_handler.get_concatenated_name(linked_fastq_name)
        return Path(
            self.root_dir,
            link_object.family.internal_id,
            "fastq",
            concatenated_fastq_name,
        )

    @staticmethod
    def get_sample_type(sample_obj: models.Sample) -> str:
        """Returns tissue type of a sample"""
        if sample_obj.is_tumour:
            return "tumor"
        return "normal"

    def get_verified_bed(self, sample_data: dict, panel_bed: Path) -> Optional[str]:
        """ "Takes a dict with samples and attributes.
        Retrieves unique attributes for application type and target_bed.
        Verifies that those attributes are the same across multiple samples,
        where applicable.
        Verifies that the attributes are valid BALSAMIC attributes
        If application type requires bed, returns path to bed.

        Raises BalsamicStartError:
        - When application type invalid for balsamic
        - When multiple samples have different parameters
        - When bed file required for analysis, but is not set or cannot be retrieved.
        """
        application_types = {v["application_type"].lower() for k, v in sample_data.items()}

        target_beds = {v["target_bed"] for k, v in sample_data.items()}

        if not application_types.issubset(self.__BALSAMIC_APPLICATIONS):
            raise BalsamicStartError("Case application not compatible with BALSAMIC")
        if len(application_types) != 1:
            raise BalsamicStartError("Multiple application types found in LIMS")
        if not application_types.issubset(self.__BALSAMIC_BED_APPLICATIONS):
            if panel_bed:
                raise BalsamicStartError("Cannot set panel_bed for WGS sample!")
            return None
        if panel_bed:
            return panel_bed.as_posix()
        if len(target_beds) == 1:
            target_bed = target_beds.pop()
            if not target_bed:
                raise BalsamicStartError(
                    f"Application type {application_types.pop()} requires a bed file to be analyzed!"
                )
            return Path(self.bed_path, target_bed).as_posix()

    @staticmethod
    def get_verified_pon(panel_bed: Path, pon_cnn: str) -> Optional[str]:
        """Returns the validated PoN path

        Raises BalsamicStartError:
            When there is a missmatch between the PoN and the panel bed file names
        """
        if pon_cnn:
            pon_cnn = Path(str(pon_cnn))
            if panel_bed.stem not in pon_cnn.stem:
                raise BalsamicStartError(
                    f"The specified PoN reference file {pon_cnn} does not match the panel bed {panel_bed}"
                )

            return pon_cnn.as_posix()

        return pon_cnn

    @staticmethod
    def get_verified_tumor_path(sample_data: dict) -> str:
        """Takes a dict with samples and attributes, and returns the path
        of tumor sample.
        If the number of paths is exactly 1, the path is returned.
        Raises BalsamicStartError:
            When there are no tumor samples, or more than one tumor sample
        """
        tumor_paths = [
            val["concatenated_path"]
            for key, val in sample_data.items()
            if val["tissue_type"] == "tumor"
        ]
        if len(tumor_paths) != 1:
            raise BalsamicStartError(
                f"Invalid number of tumor samples: {len(tumor_paths)}, "
                f"BALSAMIC analysis requires exactly 1 tumor sample per case to run successfully!"
            )
        return tumor_paths[0]

    @staticmethod
    def get_verified_normal_path(sample_data: dict) -> Optional[str]:
        """Takes a dict with samples and attributes, and retrieves the path
        of normal sample.
        If the number of paths is exactly 1, the path is returned.
        If there are no paths, then the sample is not paired, and None is returned.
        Raises BalsamicStartError:
            When there is more than one normal sample.
        """
        normal_paths = [
            val["concatenated_path"]
            for key, val in sample_data.items()
            if val["tissue_type"] == "normal"
        ]
        if len(normal_paths) > 1:
            raise BalsamicStartError(
                f"Invalid number of normal samples: {len(normal_paths)}, only up to 1 allowed!!"
            )
        if not normal_paths:
            return None
        return normal_paths[0]

    def get_latest_raw_file_data(self, case_id: str, tags: list) -> Any:
        """Retrieves the data of the latest file associated to a specific case ID and a list of tags"""

        version = self.housekeeper_api.last_version(bundle=case_id)
        raw_file = self.housekeeper_api.get_files(
            bundle=case_id, version=version.id, tags=tags
        ).first()

        if not raw_file:
            raise FileNotFoundError(
                f"No file associated to {tags} was found in housekeeper for {case_id}"
            )

        with open(Path(raw_file.full_path), "r") as stream:
            data = yaml.safe_load(stream)

        return data

    def get_latest_metadata(self, case_id: str) -> BalsamicAnalysis:
        """Get the latest metadata of a specific BALSAMIC case"""

        config_raw_data = self.get_latest_raw_file_data(case_id, [BalsamicAnalysisTag.CONFIG])
        metrics_raw_data = self.get_latest_raw_file_data(case_id, [BalsamicAnalysisTag.QC_METRICS])

        if config_raw_data and metrics_raw_data:
            try:
                balsamic_analysis = self.parse_analysis(
                    config_raw=config_raw_data, qc_metrics_raw=metrics_raw_data
                )
                return balsamic_analysis
            except ValidationError as error:
                LOG.error(
                    "get_latest_metadata failed for '%s', missing attribute: %s",
                    case_id,
                    error,
                )
                raise error
        else:
            LOG.error(f"Unable to retrieve the latest metadata for {case_id}")
            raise CgError

    def parse_analysis(self, config_raw: dict, qc_metrics_raw: dict, **kwargs) -> BalsamicAnalysis:
        """Returns a formatted BalsamicAnalysis object"""

        sequencing_type = config_raw["analysis"]["sequencing_type"]
        qc_metrics = dict()

        for value in qc_metrics_raw:
            sample_metric = BalsamicMetricsBase(**value)
            try:
                qc_metrics[sample_metric.id].update(
                    {sample_metric.name.lower(): sample_metric.value}
                )
            except KeyError:
                qc_metrics[sample_metric.id] = {sample_metric.name.lower(): sample_metric.value}

        return BalsamicAnalysis(
            config=config_raw,
            sample_metrics=self.cast_metrics_type(sequencing_type, qc_metrics),
        )

    @staticmethod
    def cast_metrics_type(
        sequencing_type: str, metrics: dict
    ) -> Union[BalsamicTargetedQCMetrics, BalsamicWGSQCMetrics]:
        """Cast metrics model type according to the sequencing type"""

        if metrics:
            for k, v in metrics.items():
                metrics[k] = (
                    BalsamicWGSQCMetrics(**v)
                    if sequencing_type == "wgs"
                    else BalsamicTargetedQCMetrics(**v)
                )

        return metrics

    def get_tumor_sample_name(self, case_id: str) -> Optional[str]:
        sample_obj = (
            self.status_db.query(models.Sample)
            .join(models.Family.links, models.FamilySample.sample)
            .filter(models.Family.internal_id == case_id)
            .filter(models.Sample.is_tumour == True)
            .first()
        )
        if sample_obj:
            return sample_obj.internal_id

    def get_normal_sample_name(self, case_id: str) -> Optional[str]:

        sample_obj = (
            self.status_db.query(models.Sample)
            .join(models.Family.links, models.FamilySample.sample)
            .filter(models.Family.internal_id == case_id)
            .filter(models.Sample.is_tumour == False)
            .first()
        )
        if sample_obj:
            return sample_obj.internal_id

    def get_verified_config_case_arguments(
        self, case_id: str, genome_version: str, panel_bed: str, pon_cnn: str
    ) -> dict:
        """Takes a dictionary with per-sample parameters,
        validates them, and transforms into command line arguments
        Raises BalsamicStartError:
            When no samples associated with case are marked for BALSAMIC analysis
        """
        sample_data = self.get_sample_params(case_id=case_id, panel_bed=panel_bed)
        if len(sample_data) == 0:
            raise BalsamicStartError(f"{case_id} has no samples tagged for BALSAMIC analysis!")
        if panel_bed:
            if Path(f"{panel_bed}").is_file():
                panel_bed = Path(f"{panel_bed}")
            else:
                derived_panel_bed = Path(
                    self.bed_path,
                    self.status_db.bed_version(panel_bed).filename,
                )
                if not derived_panel_bed.is_file():
                    raise BalsamicStartError(
                        f"{panel_bed} or {derived_panel_bed} are not valid paths to a BED file. "
                        f"Please provide absolute path to desired BED file or a valid bed shortname!"
                    )
                panel_bed = derived_panel_bed

        return {
            "case_id": case_id,
            "genome_version": genome_version,
            "normal": self.get_verified_normal_path(sample_data=sample_data),
            "tumor": self.get_verified_tumor_path(sample_data=sample_data),
            "panel_bed": self.get_verified_bed(sample_data=sample_data, panel_bed=panel_bed),
            "pon_cnn": self.get_verified_pon(pon_cnn=pon_cnn, panel_bed=panel_bed),
            "tumor_sample_name": self.get_tumor_sample_name(case_id=case_id),
            "normal_sample_name": self.get_normal_sample_name(case_id=case_id),
        }

    def build_sample_id_map_string(self, case_id: str) -> str:
        """Creates sample info string for balsamic with format lims_id:tumor/normal:customer_sample_id"""

        tumor_sample_lims_id = self.get_tumor_sample_name(case_id=case_id)
        tumor_string = f"{tumor_sample_lims_id}:tumor:{self.status_db.sample(internal_id=tumor_sample_lims_id).name}"
        normal_sample_lims_id = self.get_normal_sample_name(case_id=case_id)
        if normal_sample_lims_id:
            normal_string = f"{normal_sample_lims_id}:normal:{self.status_db.sample(internal_id=normal_sample_lims_id).name}"
            return ",".join([tumor_string, normal_string])
        return tumor_string

    def build_case_id_map_string(self, case_id: str) -> Optional[str]:
        """Creates case info string for balsamic with format panel_shortname:case_name:application_tag"""

        case_obj: models.Family = self.status_db.family(case_id)
        sample_obj: models.Sample = case_obj.links[0].sample
        if sample_obj.from_sample:
            sample_obj: models.Sample = self.status_db.sample(sample_obj.from_sample)
        capture_kit = self.lims_api.capture_kit(sample_obj.internal_id)
        if capture_kit:
            panel_shortname = self.status_db.bed_version(capture_kit).shortname
        elif self.get_application_type(case_obj.links[0].sample) == "wgs":
            panel_shortname = "Whole_Genome"
        else:
            return
        application_tag = (
            self.status_db.query(models.ApplicationVersion)
            .filter(models.ApplicationVersion.id == case_obj.links[0].sample.application_version_id)
            .first()
            .application.tag
        )
        return f"{panel_shortname}:{case_obj.name}:{application_tag}"

    @staticmethod
    def print_sample_params(case_id: str, sample_data: dict) -> None:
        """Outputs a table of samples to be displayed in log"""

        LOG.info(f"Case {case_id} has following BALSAMIC samples:")
        LOG.info(
            "{:<20} {:<20} {:<20} {:<20}".format(
                "SAMPLE ID", "TISSUE TYPE", "APPLICATION", "BED VERSION"
            )
        )
        for key in sample_data:
            LOG.info(
                "{:<20} {:<20} {:<20} {:<20}".format(
                    key,
                    str(sample_data[key]["tissue_type"]),
                    str(sample_data[key]["application_type"]),
                    str(sample_data[key]["target_bed"]),
                )
            )
        LOG.info("")

    def get_sample_params(self, case_id: str, panel_bed: Optional[str]) -> dict:

        """Returns a dictionary of attributes for each sample in given family,
        where SAMPLE ID is used as key"""

        sample_data = {
            link_object.sample.internal_id: {
                "tissue_type": self.get_sample_type(link_object.sample),
                "concatenated_path": self.get_concatenated_fastq_path(link_object).as_posix(),
                "application_type": self.get_application_type(link_object.sample),
                "target_bed": self.resolve_target_bed(panel_bed=panel_bed, link_object=link_object),
            }
            for link_object in self.status_db.family(case_id).links
        }

        self.print_sample_params(case_id=case_id, sample_data=sample_data)
        return sample_data

    def get_case_application_type(self, case_id: str) -> str:
        application_types = {
            self.get_application_type(link_object.sample)
            for link_object in self.status_db.family(case_id).links
        }

        if application_types:
            return application_types.pop().lower()

    def resolve_target_bed(
        self, panel_bed: Optional[str], link_object: models.FamilySample
    ) -> Optional[str]:
        if panel_bed:
            return panel_bed
        if self.get_application_type(link_object.sample) not in self.__BALSAMIC_BED_APPLICATIONS:
            return None
        return self.get_target_bed_from_lims(link_object.family.internal_id)

    def get_pipeline_version(self, case_id: str) -> str:
        LOG.debug("Fetch pipeline version")
        sample_config = self.get_case_config_path(case_id=case_id)
        config_data: dict = json.load(open(sample_config, "r"))
        return config_data["analysis"]["BALSAMIC_version"]

    def family_has_correct_number_tumor_normal_samples(self, case_id: str) -> bool:
        """Evaluates if a case has exactly one tumor and up to one normal sample in ClinicalDB.
        This check is only applied to filter jobs which start automatically"""
        query = (
            self.status_db.query(models.Sample)
            .join(models.Family.links, models.FamilySample.sample)
            .filter(models.Family.internal_id == case_id)
            .filter(models.Family.data_analysis == self.pipeline)
        )
        return all(
            [
                len(query.filter(models.Sample.is_tumour == False).all()) <= 1,
                len(query.filter(models.Sample.is_tumour == True).all()) == 1,
            ]
        )

    def get_valid_cases_to_analyze(self) -> list:
        """Retrieve a list of balsamic cases without analysis,
        where samples have enough reads to be analyzed"""

        return [
            case_object.internal_id
            for case_object in self.get_cases_to_analyze()
            if self.family_has_correct_number_tumor_normal_samples(case_object.internal_id)
        ]

    @staticmethod
    def __build_command_str(options: dict) -> List[str]:
        formatted_options = []
        for key, val in options.items():
            if val:
                formatted_options.append(str(key))
                formatted_options.append(str(val))
        return formatted_options

    def config_case(
        self,
        case_id: str,
        genome_version: str,
        panel_bed: str,
        pon_cnn: str,
        dry_run: bool = False,
    ) -> None:
        """Create config file for BALSAMIC analysis"""
        arguments = self.get_verified_config_case_arguments(
            case_id=case_id,
            genome_version=genome_version,
            panel_bed=panel_bed,
            pon_cnn=pon_cnn,
        )
        command = ["config", "case"]
        options = self.__build_command_str(
            {
                "--analysis-dir": self.root_dir,
                "--balsamic-cache": self.balsamic_cache,
                "--case-id": arguments.get("case_id"),
                "--genome-version": arguments.get("genome_version"),
                "--normal": arguments.get("normal"),
                "--tumor": arguments.get("tumor"),
                "--panel-bed": arguments.get("panel_bed"),
                "--pon-cnn": arguments.get("pon_cnn"),
                "--umi-trim-length": arguments.get("umi_trim_length"),
                "--tumor-sample-name": arguments.get("tumor_sample_name"),
                "--normal-sample-name": arguments.get("normal_sample_name"),
            }
        )
        parameters = command + options
        self.process.run_command(parameters=parameters, dry_run=dry_run)

    def run_analysis(
        self,
        case_id: str,
        analysis_type: Optional[str],
        run_analysis: bool = True,
        slurm_quality_of_service: Optional[str] = None,
        dry_run: bool = False,
    ) -> None:
        """Execute BALSAMIC run analysis with given options"""

        command = ["run", "analysis"]
        run_analysis = ["--run-analysis"] if run_analysis else []
        benchmark = ["--benchmark"]
        options = self.__build_command_str(
            {
                "--account": self.account,
                "--mail-user": self.email,
                "--qos": slurm_quality_of_service or self.get_slurm_qos_for_case(case_id=case_id),
                "--sample-config": self.get_case_config_path(case_id=case_id),
                "--analysis-type": analysis_type or self.get_analysis_type(case_id),
            }
        )
        parameters = command + options + run_analysis + benchmark
        self.process.run_command(parameters=parameters, dry_run=dry_run)

    def report_deliver(
        self, case_id: str, analysis_type: Optional[str] = None, dry_run: bool = False
    ) -> None:
        """Execute BALSAMIC report deliver with given options"""

        command = ["report", "deliver"]
        options = self.__build_command_str(
            {
                "--sample-config": self.get_case_config_path(case_id=case_id),
                "--analysis-type": analysis_type or self.get_analysis_type(case_id),
            }
        )
        parameters = command + options
        self.process.run_command(parameters=parameters, dry_run=dry_run)

    def get_analysis_type(self, case_id: str) -> Optional[str]:
        case_obj = self.status_db.family(case_id)
        if case_obj.data_delivery in [DataDelivery.FASTQ_QC, DataDelivery.FASTQ]:
            return "qc"



"""Tests the cli for generating delivery reports"""

import logging
from datetime import datetime

from cg.cli.generate.commands import delivery_report
from cg.constants import EXIT_SUCCESS, EXIT_FAIL


def test_delivery_report_invalid_case(mip_dna_context, cli_runner, caplog):
    """Tests the delivery report command for an invalid case"""

    caplog.set_level(logging.INFO)

    # GIVEN a MIP DNA context object

    # WHEN calling delivery_report with an invalid case name
    result = cli_runner.invoke(
        delivery_report,
        ["not a case"],
        obj=mip_dna_context,
    )

    # THEN the command should fail due to an invalid case ID
    assert "Invalid case ID. Retrieving cases without a delivery report." in caplog.text
    assert "There are no cases available to generate delivery reports" in result.output
    assert result.exit_code == EXIT_FAIL


def test_delivery_report_dry_run(mip_dna_context, cli_runner, case_id, caplog):
    """Tests the delivery report command with a dry run option"""

    caplog.set_level(logging.INFO)

    # GIVEN a MIP DNA context object

    # WHEN calling delivery_report with a dry option
    result = cli_runner.invoke(
        delivery_report,
        [case_id, "--dry-run"],
        obj=mip_dna_context,
    )

    # THEN the command should be invoked successfully
    assert "create_delivery_report" in caplog.text
    assert "create_delivery_report_file" not in caplog.text
    assert "case=" + case_id in caplog.text
    assert "analysis_date=" + str(datetime.now().date()) in caplog.text
    assert result.exit_code == EXIT_SUCCESS


def test_delivery_report(mip_dna_context, cli_runner, case_id, caplog):
    """Tests the delivery report command expecting a rendered html file"""

    caplog.set_level(logging.INFO)

    # GIVEN a MIP DNA context object

    # WHEN calling delivery_report with a dry option
    cli_runner.invoke(
        delivery_report,
        [case_id],
        obj=mip_dna_context,
    )

    # THEN the report html should have been created and added to HK
    assert "create_delivery_report_file" in caplog.text
    assert "add_delivery_report_to_hk" in caplog.text



"""Module for deliver and rsync customer inbox on hasta to customer inbox on caesar"""
import datetime as dt
import logging
from pathlib import Path
from typing import List, Optional

from cg.apps.cgstats.db.models import Version
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.constants import HK_FASTQ_TAGS
from cg.meta.meta import MetaAPI
from cg.meta.rsync.sbatch import ERROR_RSYNC_FUNCTION, RSYNC_CONTENTS_COMMAND
from cg.models.cg_config import CGConfig
from cg.models.slurm.sbatch import Sbatch
from cg.store import models
from cg.utils.checksum.checksum import check_md5sum, extract_md5sum

LOG = logging.getLogger(__name__)


class ExternalDataAPI(MetaAPI):
    def __init__(self, config: CGConfig):
        super().__init__(config)
        self.destination_path: str = config.external.hasta
        self.source_path: str = config.external.caesar
        self.base_path: str = config.data_delivery.base_path
        self.account: str = config.data_delivery.account
        self.mail_user: str = config.data_delivery.mail_user
        self.slurm_api: SlurmAPI = SlurmAPI()
        self.RSYNC_FILE_POSTFIX: str = "_rsync_external_data"

    def create_log_dir(self, dry_run: bool, ticket_id: int) -> Path:
        """Creates a directory for log file to be stored"""
        timestamp: dt.datetime = dt.datetime.now()
        timestamp_str: str = timestamp.strftime("%y%m%d_%H_%M_%S_%f")
        folder_name: Path = Path("_".join([str(ticket_id), timestamp_str]))
        log_dir: Path = Path(self.base_path, folder_name)
        LOG.info("Creating folder: %s", log_dir)
        if dry_run:
            LOG.info("Would have created path %s, but this is a dry run", log_dir)
            return log_dir
        log_dir.mkdir(parents=True, exist_ok=False)
        return log_dir

    def get_source_path(
        self,
        customer: str,
        ticket_id: int,
        cust_sample_id: Optional[str] = "",
    ) -> Path:
        """Returns the path to where the sample files are fetched from"""
        return Path(self.source_path % customer, str(ticket_id), cust_sample_id)

    def get_destination_path(self, customer: str, lims_sample_id: Optional[str] = "") -> Path:
        """Returns the path to where the files are to be transferred"""
        return Path(self.destination_path % customer, lims_sample_id)

    def transfer_sample_files_from_source(self, dry_run: bool, ticket_id: int) -> None:
        """Transfers all sample files, related to given ticket, from source to destination"""
        cust: str = self.status_db.get_customer_id_from_ticket(ticket_id=ticket_id)
        log_dir: Path = self.create_log_dir(ticket_id=ticket_id, dry_run=dry_run)
        error_function: str = ERROR_RSYNC_FUNCTION.format()
        Path(self.destination_path % cust).mkdir(exist_ok=True)

        command: str = RSYNC_CONTENTS_COMMAND.format(
            source_path=self.get_source_path(customer=cust, ticket_id=ticket_id),
            destination_path=self.get_destination_path(customer=cust),
        )
        sbatch_parameters: Sbatch = Sbatch(
            job_name=str(ticket_id) + self.RSYNC_FILE_POSTFIX,
            account=self.account,
            number_tasks=1,
            memory=1,
            log_dir=str(log_dir),
            email=self.mail_user,
            hours=24,
            commands=command,
            error=error_function,
        )
        self.slurm_api.set_dry_run(dry_run=dry_run)
        sbatch_content: str = self.slurm_api.generate_sbatch_content(
            sbatch_parameters=sbatch_parameters
        )
        sbatch_path: Path = Path(log_dir, str(ticket_id) + self.RSYNC_FILE_POSTFIX + ".sh")
        self.slurm_api.submit_sbatch(sbatch_content=sbatch_content, sbatch_path=sbatch_path)
        LOG.info(
            "The folder {src_path} is now being rsynced to hasta".format(
                src_path=self.get_source_path(customer=cust, ticket_id=ticket_id)
            )
        )

    def get_all_fastq(self, sample_folder: Path) -> List[Path]:
        """Returns a list of all fastq.gz files in given folder"""
        all_fastqs: List[Path] = []
        for leaf in sample_folder.glob("*fastq.gz"):
            abs_path: Path = sample_folder.joinpath(leaf)
            LOG.info("Found file %s inside folder %s" % (str(abs_path), sample_folder))
            all_fastqs.append(abs_path)
        return all_fastqs

    def get_all_paths(self, customer: str, lims_sample_id: str) -> List[Path]:
        """Returns the paths of all fastq files associated to the sample"""
        fastq_folder: Path = self.get_destination_path(
            lims_sample_id=lims_sample_id, customer=customer
        )
        all_fastq_in_folder: List[Path] = self.get_all_fastq(sample_folder=fastq_folder)
        return all_fastq_in_folder

    def check_fastq_md5sum(self, fastq_path) -> Optional[Path]:
        """Returns the path of the input file if it does not match its md5sum"""
        if Path(str(fastq_path) + ".md5").exists():
            given_md5sum: str = extract_md5sum(md5sum_file=Path(str(fastq_path) + ".md5"))
            if not check_md5sum(file_path=fastq_path, md5sum=given_md5sum):
                return fastq_path

    def get_available_samples(self, folder: Path, ticket_id: int) -> List[models.Sample]:
        """Returns the samples from given ticket that are present in the provided folder"""
        available_folders: List[str] = [sample_path.parts[-1] for sample_path in folder.iterdir()]
        available_samples: List[models.Sample] = [
            sample
            for sample in self.status_db.get_samples_from_ticket(ticket_id=ticket_id)
            if sample.internal_id in available_folders or sample.name in available_folders
        ]
        return available_samples

    def add_files_to_bundles(
        self, fastq_paths: List[Path], last_version: Version, lims_sample_id: str
    ):
        """Adds the given fastq files to the the hk-bundle"""
        for path in fastq_paths:
            LOG.info("Adding path %s to bundle %s in housekeeper" % (path, lims_sample_id))
            self.housekeeper_api.add_file(path=path, version_obj=last_version, tags=HK_FASTQ_TAGS)

    def get_failed_fastq_paths(self, fastq_paths_to_add: List[Path]) -> List[Path]:
        failed_sum_paths: List[Path] = []
        for path in fastq_paths_to_add:
            failed_path: Optional[Path] = self.check_fastq_md5sum(path)
            if failed_path:
                failed_sum_paths.append(failed_path)
        return failed_sum_paths

    def get_fastq_paths_to_add(
        self, customer: str, hk_version: Version, lims_sample_id: str
    ) -> List[Path]:
        paths: List[Path] = self.get_all_paths(lims_sample_id=lims_sample_id, customer=customer)
        fastq_paths_to_add: List[Path] = self.housekeeper_api.check_bundle_files(
            file_paths=paths,
            bundle_name=lims_sample_id,
            last_version=hk_version,
            tags=HK_FASTQ_TAGS,
        )
        return fastq_paths_to_add

    def curate_sample_folder(self, cust_name: str, force: bool, sample_folder: Path) -> None:
        """Changes the name of the folder to the internal_id. If force is true replaces any previous folder"""
        customer: models.Customer = self.status_db.customer(internal_id=cust_name)
        customer_folder: Path = sample_folder.parent
        sample: models.Sample = self.status_db.find_samples(
            customer=customer, name=sample_folder.name
        ).first()
        if (sample and not customer_folder.joinpath(sample.internal_id).exists()) or (
            sample and force
        ):
            sample_folder.rename(customer_folder.joinpath(sample.internal_id))
        elif not sample and not self.status_db.sample(sample_folder.name):
            raise Exception(
                f"{sample_folder} is not a sample present in statusdb. Move or remove it to continue"
            )

    def add_transfer_to_housekeeper(
        self, ticket_id: int, dry_run: bool = False, force: bool = False
    ) -> None:
        """Creates sample bundles in housekeeper and adds the available files corresponding to the ticket to the
        bundle"""
        failed_paths: List[Path] = []
        cust: str = self.status_db.get_customer_id_from_ticket(ticket_id=ticket_id)
        destination_folder_path: Path = self.get_destination_path(customer=cust)
        for sample_folder in destination_folder_path.iterdir():
            self.curate_sample_folder(cust_name=cust, sample_folder=sample_folder, force=force)
        available_samples: List[models.Sample] = self.get_available_samples(
            folder=destination_folder_path, ticket_id=ticket_id
        )
        cases_to_start: List[dict] = []
        for sample in available_samples:
            cases_to_start.extend(
                self.status_db.cases(sample_id=sample.internal_id, exclude_analysed=True)
            )
            last_version: Version = self.housekeeper_api.get_create_version(
                bundle=sample.internal_id
            )
            fastq_paths_to_add: List[Path] = self.get_fastq_paths_to_add(
                customer=cust, hk_version=last_version, lims_sample_id=sample.internal_id
            )
            self.add_files_to_bundles(
                fastq_paths=fastq_paths_to_add,
                last_version=last_version,
                lims_sample_id=sample.internal_id,
            )
            failed_paths.extend(self.get_failed_fastq_paths(fastq_paths_to_add=fastq_paths_to_add))
        if failed_paths:
            LOG.info(
                "The following samples did not match the given md5sum: "
                + ", ".join([str(path) for path in failed_paths])
            )
            LOG.info("Changes in housekeeper will not be committed and no cases will be added")
            return
        if dry_run:
            LOG.info("No changes will be committed since this is a dry-run")
            return
        self.housekeeper_api.commit()
        for case in cases_to_start:
            self.status_db.set_case_action(case_id=case["internal_id"], action="analyze")



import logging

from cgmodels.cg.constants import Pipeline

from cg.cli.upload.clinical_delivery import upload_fastq, auto_fastq
from cg.store import models


def test_auto_fastq_not_started(
    analysis_obj: models.Analysis, caplog, fastq_context, cli_runner, case_id: str
):
    """Tests if the command finds a non-uploaded analysis and attempts to start it"""
    caplog.set_level(logging.INFO)
    # GIVEN a case to be delivered
    analysis_obj.pipeline = Pipeline.FASTQ
    fastq_context.status_db.commit()
    fastq_context.status_db.session.close()
    # WHEN the upload command is invoked with dry run
    cli_runner.invoke(auto_fastq, ["--dry-run"], obj=fastq_context)

    # THEN the content of the .sh file should be in the caplog
    assert f"#SBATCH --job-name={case_id}_rsync" in caplog.text


def test_upload_fastq(caplog, fastq_context, cli_runner, case_id: str):
    """Test if the function successfully creates an sbatch script"""
    caplog.set_level(logging.INFO)
    # GIVEN a case to be delivered

    # WHEN the upload command is invoked with dry run
    cli_runner.invoke(upload_fastq, [case_id, "--dry-run"], obj=fastq_context)

    # THEN the content of the .sh file should be in the caplog
    assert f"#SBATCH --job-name={case_id}_rsync" in caplog.text



"""Fixtures for the upload scout api tests"""

import json
import logging
from datetime import datetime
from pathlib import Path
from typing import List

import pytest
from cg.constants import Pipeline, DataDelivery
from cg.meta.upload.scout.balsamic_config_builder import BalsamicConfigBuilder
from cg.meta.upload.scout.mip_config_builder import MipConfigBuilder
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.models.scout.scout_load_config import MipLoadConfig
from cg.store import Store, models
from housekeeper.store import models as hk_models

# Mocks
from tests.mocks.hk_mock import MockHousekeeperAPI
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.madeline import MockMadelineAPI
from tests.mocks.scout import MockScoutAPI
from tests.store_helpers import StoreHelpers
from tests.mocks.mip_analysis_mock import MockMipAnalysis

LOG = logging.getLogger(__name__)


@pytest.fixture(name="rna_case_id")
def fixture_rna_case_id() -> str:
    """Return a rna case id"""
    return "affirmativecat"


@pytest.fixture(name="dna_case_id")
def fixture_dna_case_id(case_id) -> str:
    """Return a dna case id"""
    return case_id


@pytest.fixture(name="rna_sample_son_id")
def fixture_rna_sample_son_id() -> str:
    """Return a rna sample id"""
    return "rna_son"


@pytest.fixture(name="rna_sample_daughter_id")
def fixture_rna_sample_daughter_id() -> str:
    """Return a rna sample id"""
    return "rna_daughter"


@pytest.fixture(name="rna_sample_mother_id")
def fixture_rna_sample_mother_id() -> str:
    """Return a rna sample id"""
    return "rna_mother"


@pytest.fixture(name="rna_sample_father_id")
def fixture_rna_sample_father_id() -> str:
    """Return a rna sample id"""
    return "rna_father"


@pytest.fixture(name="dna_sample_son_id")
def fixture_dna_sample_son_id() -> str:
    """Return a rna sample id"""
    return "dna_son"


@pytest.fixture(name="dna_sample_daughter_id")
def fixture_dna_sample_daughter_id() -> str:
    """Return a rna sample id"""
    return "dna_daughter"


@pytest.fixture(name="dna_sample_mother_id")
def fixture_dna_sample_mother_id() -> str:
    """Return a rna sample id"""
    return "dna_mother"


@pytest.fixture(name="dna_sample_father_id")
def fixture_dna_sample_father_id() -> str:
    """Return a rna sample id"""
    return "dna_father"


@pytest.fixture(name="rna_store")
def fixture_rna_store(
    base_store: Store,
    helpers: StoreHelpers,
    rna_case_id: str,
    dna_case_id: str,
) -> Store:
    """Populate store with an rna case that is connected to a dna case via sample.subject_id"""

    store: Store = base_store

    # an existing RNA case with related sample
    rna_case = helpers.ensure_case(
        store=store,
        name="rna_case",
        customer=helpers.ensure_customer(store=store),
        data_analysis=Pipeline.MIP_RNA,
        data_delivery=DataDelivery.SCOUT,
    )
    rna_case.internal_id = rna_case_id

    rna_sample_son = helpers.add_sample(store=store, name="rna_son", subject_id="son")
    rna_sample_daughter = helpers.add_sample(
        store=store, name="rna_daughter", subject_id="daughter"
    )
    rna_sample_mother = helpers.add_sample(store=store, name="rna_mother", subject_id="mother")
    rna_sample_father = helpers.add_sample(store=store, name="rna_father", subject_id="father")
    helpers.add_relationship(
        store=store,
        sample=rna_sample_son,
        case=rna_case,
        mother=rna_sample_mother,
        father=rna_sample_father,
        status="affected",
    )
    helpers.add_relationship(
        store=store,
        sample=rna_sample_daughter,
        case=rna_case,
        mother=rna_sample_mother,
        father=rna_sample_father,
        status="unaffected",
    )
    helpers.add_relationship(
        store=store, sample=rna_sample_mother, case=rna_case, status="unaffected"
    )
    helpers.add_relationship(
        store=store, sample=rna_sample_father, case=rna_case, status="affected"
    )

    for link in rna_case.links:
        link.sample.internal_id = link.sample.name

    # an existing DNA case with related sample
    dna_case = helpers.ensure_case(
        store=store,
        name="dna_case",
        customer=helpers.ensure_customer(store=store),
        data_analysis=Pipeline.MIP_DNA,
        data_delivery=DataDelivery.SCOUT,
    )
    dna_case.internal_id = dna_case_id

    dna_sample_son = helpers.add_sample(store=store, name="dna_son", subject_id="son")
    dna_sample_daughter = helpers.add_sample(
        store=store, name="dna_daughter", subject_id="daughter"
    )
    dna_sample_mother = helpers.add_sample(store=store, name="dna_mother", subject_id="mother")
    dna_sample_father = helpers.add_sample(store=store, name="dna_father", subject_id="father")
    helpers.add_relationship(
        store=store,
        sample=dna_sample_son,
        case=dna_case,
        mother=dna_sample_mother,
        father=dna_sample_father,
        status="affected",
    )
    helpers.add_relationship(
        store=store,
        sample=dna_sample_daughter,
        case=dna_case,
        mother=dna_sample_mother,
        father=dna_sample_father,
        status="unaffected",
    )
    helpers.add_relationship(
        store=store, sample=dna_sample_mother, case=dna_case, status="unaffected"
    )
    helpers.add_relationship(
        store=store, sample=dna_sample_father, case=dna_case, status="affected"
    )

    for link in dna_case.links:
        link.sample.internal_id = link.sample.name

    store.commit()
    return store


@pytest.fixture(name="lims_family")
def fixture_lims_family() -> dict:
    """Returns a lims-like case of samples"""
    return json.load(open("tests/fixtures/report/lims_family.json"))


@pytest.fixture(name="lims_samples")
def fixture_lims_samples(lims_family: dict) -> List[dict]:
    """Returns the samples of a lims case"""
    return lims_family["samples"]


@pytest.fixture(scope="function", name="mip_dna_analysis_hk_bundle_data")
def fixture_mip_dna_analysis_hk_bundle_data(
    case_id: str, timestamp: datetime, mip_dna_analysis_dir: Path, sample_id: str
) -> dict:
    """Get some bundle data for housekeeper"""
    data = {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(mip_dna_analysis_dir / "snv.vcf"),
                "archive": False,
                "tags": ["vcf-snv-clinical"],
            },
            {
                "path": str(mip_dna_analysis_dir / "sv.vcf"),
                "archive": False,
                "tags": ["vcf-sv-clinical"],
            },
            {
                "path": str(mip_dna_analysis_dir / "snv_research.vcf"),
                "archive": False,
                "tags": ["vcf-snv-research"],
            },
            {
                "path": str(mip_dna_analysis_dir / "sv_research.vcf"),
                "archive": False,
                "tags": ["vcf-sv-research"],
            },
            {
                "path": str(mip_dna_analysis_dir / "str.vcf"),
                "archive": False,
                "tags": ["vcf-str"],
            },
            {
                "path": str(mip_dna_analysis_dir / "smn.vcf"),
                "archive": False,
                "tags": ["smn-calling"],
            },
            {
                "path": str(mip_dna_analysis_dir / "adm1.cram"),
                "archive": False,
                "tags": ["cram", sample_id],
            },
            {
                "path": str(mip_dna_analysis_dir / "report.pdf"),
                "archive": False,
                "tags": ["delivery-report"],
            },
            {
                "path": str(mip_dna_analysis_dir / "adm1.mt.bam"),
                "archive": False,
                "tags": ["bam-mt", sample_id],
            },
            {
                "path": str(mip_dna_analysis_dir / "vcf2cytosure.txt"),
                "archive": False,
                "tags": ["vcf2cytosure", sample_id],
            },
            {
                "path": str(mip_dna_analysis_dir / "multiqc.html"),
                "archive": False,
                "tags": ["multiqc-html", sample_id],
            },
        ],
    }
    return data


@pytest.fixture(scope="function", name="mip_rna_analysis_hk_bundle_data")
def fixture_mip_rna_analysis_hk_bundle_data(
    rna_case_id: str,
    timestamp: datetime,
    mip_dna_analysis_dir: Path,
    rna_sample_son_id: str,
    rna_sample_daughter_id: str,
    rna_sample_mother_id: str,
    rna_sample_father_id: str,
) -> dict:
    """Get some bundle data for housekeeper"""

    files: [dict] = [
        {
            "path": str(mip_dna_analysis_dir / f"{rna_case_id}_report.selected.pdf"),
            "archive": False,
            "tags": ["fusion", "pdf", "clinical", rna_case_id],
        },
        {
            "path": str(mip_dna_analysis_dir / f"{rna_case_id}_report.pdf"),
            "archive": False,
            "tags": ["fusion", "pdf", "research", rna_case_id],
        },
    ]
    for sample_id in [
        rna_sample_son_id,
        rna_sample_daughter_id,
        rna_sample_mother_id,
        rna_sample_father_id,
    ]:
        files.extend(
            [
                {
                    "path": str(
                        mip_dna_analysis_dir / f"{sample_id}_lanes_1_star_sorted_sj.bigWig"
                    ),
                    "archive": False,
                    "tags": ["coverage", "bigwig", "scout", sample_id],
                },
                {
                    "path": str(
                        mip_dna_analysis_dir / f"{sample_id}_lanes_1234_star_sorted_sj.bed.gz.tbi"
                    ),
                    "archive": False,
                    "tags": ["bed", "scout", "junction", sample_id],
                },
            ]
        )

    data = {
        "name": rna_case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": files,
    }
    return data


@pytest.fixture(scope="function", name="balsamic_analysis_hk_bundle_data")
def fixture_balsamic_analysis_hk_bundle_data(
    case_id: str, timestamp: datetime, balsamic_wgs_analysis_dir: Path, sample_id: str
) -> dict:
    """Get some bundle data for housekeeper"""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(balsamic_wgs_analysis_dir / "snv.vcf"),
                "archive": False,
                "tags": ["vcf-snv-clinical"],
            },
            {
                "path": str(balsamic_wgs_analysis_dir / "sv.vcf"),
                "archive": False,
                "tags": ["vcf-sv-clinical"],
            },
            {
                "path": str(balsamic_wgs_analysis_dir / "adm1.cram"),
                "archive": False,
                "tags": ["cram", sample_id],
            },
            {
                "path": str(balsamic_wgs_analysis_dir / "ascat.output.pdf"),
                "archive": False,
                "tags": ["ascatngs", "visualization", sample_id],
            },
            {
                "path": str(balsamic_wgs_analysis_dir / "coverage_qc_report.pdf"),
                "archive": False,
                "tags": ["delivery-report"],
            },
        ],
    }


@pytest.fixture(name="balsamic_analysis_hk_version")
def fixture_balsamic_analysis_hk_version(
    housekeeper_api: MockHousekeeperAPI, balsamic_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    return helpers.ensure_hk_version(housekeeper_api, balsamic_analysis_hk_bundle_data)


@pytest.fixture(name="mip_dna_analysis_hk_version")
def fixture_mip_dna_analysis_hk_version(
    housekeeper_api: MockHousekeeperAPI, mip_dna_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    return helpers.ensure_hk_version(housekeeper_api, mip_dna_analysis_hk_bundle_data)


@pytest.fixture(name="mip_dna_analysis_hk_api")
def fixture_mip_dna_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, mip_dna_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a housekeeper api populated with some mip dna analysis files"""
    helpers.ensure_hk_version(housekeeper_api, mip_dna_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="mip_rna_analysis_hk_api")
def fixture_mip_rna_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, mip_rna_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a housekeeper api populated with some mip rna analysis files"""
    helpers.ensure_hk_version(housekeeper_api, mip_rna_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="balsamic_analysis_hk_api")
def fixture_balsamic_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, balsamic_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a housekeeper api populated with some mip analysis files"""
    helpers.ensure_hk_version(housekeeper_api, balsamic_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="mip_file_handler")
def fixture_mip_file_handler(mip_dna_analysis_hk_version: hk_models.Version) -> MipConfigBuilder:
    return MipConfigBuilder(hk_version_obj=mip_dna_analysis_hk_version)


@pytest.fixture(name="mip_dna_analysis_obj")
def fixture_mip_dna_analysis_obj(
    analysis_store_trio: Store, case_id: str, timestamp: datetime, helpers: StoreHelpers
) -> models.Analysis:
    helpers.add_synopsis_to_case(store=analysis_store_trio, case_id=case_id)
    case_obj: models.Family = analysis_store_trio.family(case_id)
    analysis_obj: models.Analysis = helpers.add_analysis(
        store=analysis_store_trio,
        case=case_obj,
        started_at=timestamp,
        pipeline=Pipeline.MIP_DNA,
        completed_at=timestamp,
    )
    for link in case_obj.links:
        helpers.add_phenotype_groups_to_sample(
            store=analysis_store_trio, sample_id=link.sample.internal_id
        )
        helpers.add_phenotype_terms_to_sample(
            store=analysis_store_trio, sample_id=link.sample.internal_id
        )
        helpers.add_subject_id_to_sample(
            store=analysis_store_trio, sample_id=link.sample.internal_id
        )
    return analysis_obj


@pytest.fixture(name="balsamic_analysis_obj")
def fixture_balsamic_analysis_obj(analysis_obj: models.Analysis) -> models.Analysis:
    analysis_obj.pipeline = Pipeline.BALSAMIC
    for link_object in analysis_obj.family.links:
        link_object.sample.application_version.application.prep_category = "wes"
        link_object.family.data_analysis = Pipeline.BALSAMIC
    return analysis_obj


@pytest.fixture(name="balsamic_umi_analysis_obj")
def fixture_balsamic_umi_analysis_obj(analysis_obj: models.Analysis) -> models.Analysis:
    analysis_obj.pipeline = Pipeline.BALSAMIC_UMI
    for link_object in analysis_obj.family.links:
        link_object.sample.application_version.application.prep_category = "wes"
        link_object.family.data_analysis = Pipeline.BALSAMIC_UMI

    return analysis_obj


@pytest.fixture(name="mip_config_builder")
def fixture_mip_config_builder(
    mip_dna_analysis_hk_version: hk_models.Version,
    mip_dna_analysis_obj: models.Analysis,
    lims_api: MockLimsAPI,
    mip_analysis_api: MockMipAnalysis,
    madeline_api: MockMadelineAPI,
) -> MipConfigBuilder:
    return MipConfigBuilder(
        hk_version_obj=mip_dna_analysis_hk_version,
        analysis_obj=mip_dna_analysis_obj,
        lims_api=lims_api,
        mip_analysis_api=mip_analysis_api,
        madeline_api=madeline_api,
    )


@pytest.fixture(name="balsamic_config_builder")
def fixture_balsamic_config_builder(
    balsamic_analysis_hk_version: hk_models.Version,
    balsamic_analysis_obj: models.Analysis,
    lims_api: MockLimsAPI,
) -> BalsamicConfigBuilder:
    return BalsamicConfigBuilder(
        hk_version_obj=balsamic_analysis_hk_version,
        analysis_obj=balsamic_analysis_obj,
        lims_api=lims_api,
    )


@pytest.fixture(name="mip_load_config")
def fixture_mip_load_config(
    mip_dna_analysis_dir: Path, case_id: str, customer_id: str
) -> MipLoadConfig:
    """Return a valid mip load_config"""
    return MipLoadConfig(
        owner=customer_id,
        family=case_id,
        vcf_snv=str(mip_dna_analysis_dir / "snv.vcf"),
        track="rare",
    )


@pytest.fixture(name="lims_api")
def fixture_lims_api(lims_samples: List[dict]) -> MockLimsAPI:
    return MockLimsAPI(samples=lims_samples)


@pytest.fixture(name="mip_analysis_api")
def fixture_mip_analysis_api() -> MockMipAnalysis:
    return MockMipAnalysis()


@pytest.fixture(name="upload_scout_api")
def fixture_upload_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    housekeeper_api: MockHousekeeperAPI,
    store: Store,
) -> UploadScoutAPI:
    """Fixture for upload_scout_api"""
    analysis_mock = MockMipAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    return UploadScoutAPI(
        hk_api=housekeeper_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )


@pytest.fixture(name="upload_mip_analysis_scout_api")
def fixture_upload_mip_analysis_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    mip_dna_analysis_hk_api: MockHousekeeperAPI,
    store: Store,
) -> UploadScoutAPI:
    """Fixture for upload_scout_api"""
    analysis_mock = MockMipAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    _api = UploadScoutAPI(
        hk_api=mip_dna_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )

    yield _api


@pytest.fixture(name="upload_balsamic_analysis_scout_api")
def fixture_upload_balsamic_analysis_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    balsamic_analysis_hk_api: MockHousekeeperAPI,
    store: Store,
) -> UploadScoutAPI:
    """Fixture for upload_scout_api"""
    analysis_mock = MockMipAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    _api = UploadScoutAPI(
        hk_api=balsamic_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
        status_db=store,
    )

    yield _api



"""Module for Balsamic Analysis API"""

import json
import logging
from pathlib import Path
from typing import List, Optional

from cg.constants import DataDelivery, Pipeline
from cg.exc import BalsamicStartError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.meta.workflow.fastq import BalsamicFastqHandler
from cg.models.cg_config import CGConfig
from cg.store import models
from cg.utils import Process

LOG = logging.getLogger(__name__)


class BalsamicUmiAnalysisAPI(BalsamicAnalysisAPI):
    """Handles communication between BALSAMIC processes
    and the rest of CG infrastructure"""

    def __init__(
        self,
        config: CGConfig,
        pipeline: Pipeline = Pipeline.BALSAMIC_UMI,
    ):
        super().__init__(config=config, pipeline=pipeline)
